#!/usr/bin/env python

import sys
import subprocess
import getopt

VERBOSITY=1
filename = ""

def call_undertaker(filename, mode = "simple"):
    """ Generates the presence conditions for `filename' in one line """
    cpppc = subprocess.Popen(['undertaker', '-q', '-j', 'coverage', '-O', 'cpp',
                              '-C', mode, filename],
                             stdin=subprocess.PIPE,
                             stdout=subprocess.PIPE,
                             stderr=subprocess.PIPE)
    cpppc.stdin.close()
    (stdout_got, stderr_got) = (cpppc.stdout.readlines(), cpppc.stderr.readlines())
    cpppc.wait()

    if 0 != cpppc.returncode or len(stderr_got) > 0:
        print "STDOUT:", stdout_got
        print "STDERR:", stderr_got
        raise RuntimeError("calling undertaker failed")

    return filter(lambda x: not x.startswith("I:"), stdout_got)

class Main:
    def __init__(self):
        self.vars = set()
        self.blocks = set()
        self.content = ""

    def cpp_results(self, cpp_args):
        """ generating preprocessor result for specific cpp command line argument"""
        flags = ['cpp'] + filter(lambda x: len(x) != 0, cpp_args.strip().split(" "))
        if VERBOSITY > 1:
            print flags

        p = subprocess.Popen(flags,
                             stdin=subprocess.PIPE,
                             stdout=subprocess.PIPE,
                             stderr=subprocess.PIPE)
        blob = p.communicate(self.content)[0]
        if 0 != p.returncode:
            raise RuntimeError("error running cpp with args: " + str(cpp_args))
        return set([ i.strip() for i in blob.split('\n') if i.startswith('B') ])


    def calc_coverage(self, mode = "simple"):
        configurations = call_undertaker(filename, mode)
        covered_set = set()
        for config in configurations:
            covered_set = covered_set.union(self.cpp_results(config))

        print "  COVERAGE %.2f%%, %d configs, %s" % (float(len(covered_set))/float(len(self.blocks)) * 100,
                                                    len(configurations), mode)
        if self.blocks != covered_set:
            print "    DIFF: " + str(list(self.blocks.difference(covered_set)))

        return float(len(covered_set))/float(len(self.blocks))

    def check(self):
        print "TEST:", filename
        return not(self.calc_coverage("simple") == self.calc_coverage("min"))


    def main(self):
        p = subprocess.Popen(['predator', '-c', filename], stdout=subprocess.PIPE)
        self.content = p.communicate()[0]
        if p.returncode != 0:
            raise RuntimeError("Error while running predator")

        continue_with_next = False

        def extract_block(line):
            if line.startswith('B'):
                self.blocks.add(line.strip())


        for line in self.content.split('\n'):
            if VERBOSITY == 2:
                print line
            extract_block(line)

        if VERBOSITY == 2:
            print self.blocks

def usage():
    print "%s [-v level] filename" % sys.argv[0]
    print "%s -h" % sys.argv[0]
    print
    print "-v: Verbosity"
    print "  -v 0: only print differences"
    print "  -v 1: print result for all combinations"
    print "  -v 2: noisy debug output"
    print
    print "Return values:"
    print " 1: Operation Error"
    print " 2: Some checks didn't pass"

if __name__ == '__main__':
    opts, args = getopt.getopt(sys.argv[1:], "hv:")

    for opt,arg in opts:
        if opt in ('-v', '--verbose'):
            VERBOSITY=int(arg)
            if VERBOSITY < 0 or VERBOSITY > 2:
                print "invalid verbosity level"
                usage()
                sys.exit(1)
        elif opt in ('-h', '--help'):
            usage()
            sys.exit(0)
        else:
            usage()
            assert False, "unknown option"

    if len(args) != 1:
        print "(only) expecting filename as argument"
        usage()
        sys.exit(1)

    filename = args[0]

    m = Main()
    m.main()
    sys.exit(2 if m.check() else 0)



#!/usr/bin/env python
#
#   vampyr - extracts presence implications from kconfig dumps
#
# Copyright (C) 2012 Reinhard Tartler <tartler@informatik.uni-erlangen.de>
# Copyright (C) 2012 Christian Dietrich <christian.dietrich@informatik.uni-erlangen.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from vamos.tools import execute, CommandFailed

import glob
import logging
import os
import re

class ExpansionError(RuntimeError):
    """ Base class of all sort of Expansion errors """
    pass


class ExpansionSanityCheckError(ExpansionError):
    """ Internal kernel config sanity checks failed, like `make silentoldconfig` """
    pass


def get_conditional_blocks(filename, autoconf_h=None, configuration_blocks=True,
                           strip_linums=True):
    """
    Counts the conditional blocks in the given source file

    The calculation is done using the 'zizler' binary from the system
    path.

    If the parameter 'autoconf_h' is set to an existing file, then the
    source file is preprocessed with 'cpp' with the given
    configuration. Usually, this will be some 'autoconf.h'. For Linux,
    this can be detected with the method find_autoconf() from the golem
    package.

    If the parameter configuration_blocks is set to true, only
    configuration control conditional blocks will be counted, otherwise
    all blocks.

    @return a non-empty list of blocks found in the source file

    """

    if configuration_blocks:
        normalizer = 'zizler -cC "%s"' % filename
    else:
        normalizer = 'zizler -c "%s"' % filename

    if autoconf_h and os.path.exists(autoconf_h):
        cmd = '%s | cpp -include %s' % (normalizer, autoconf_h)
    else:
        cmd = normalizer
    (blocks, rc) = execute(cmd, echo=True, failok=True)

    blocks = filter(lambda x: len(x) != 0 and x.startswith("B"), blocks)
    # With never versions of zizler line numbers for each block are
    # also printed. By default they are stripped, to retain backward
    # compatibility
    #  "B00 23" -> "B00"
    if strip_linums and len(blocks) > 0 and " " in blocks[0]:
        blocks = [x.split(" ",1)[0] for x in blocks]
    if rc != 0:
        logging.warning("'%s' signals exitcode: %d", cmd, rc)
        if rc == 127:
            raise CommandFailed(cmd, 127)
    return blocks


def get_block_to_linum_map(filename, configuration_blocks=False):
    """Returns a map from configuration controlled block name to a
    line count within the block"""

    blocks = get_conditional_blocks(filename, autoconf_h=False,
                                    configuration_blocks=configuration_blocks,
                                    strip_linums=False)

    d = dict([tuple(x.split(" ")) for x in blocks])
    for key in d:
        d[key] = int(d[key])
    return d


def get_loc_coverage(filename, autoconf_h=None):
    """
    Returns LOC of the given file taking the current configuration into account

    If the parameter 'autoconf_h' is set to an existing file, then the
    source file is preprocessed with 'cpp' with the given
    configuration. Usually, this will be some 'autoconf.h'. For Linux,
    configuration. Usually, this will be some 'autoconf.h'. If it is not
    set, then the file will not be preprocessed at all.

    The given filename is stripped from '#include' directives, and
    (optionally) configured with a configuration file.

    ..note: For Linux, use the method 'find_autoconf()' from the
            vamos.golem.kbuild package to find a suitable autoconf_h

    ..note: Use '/dev/null' for an empty configuration

    """

    assert(os.path.exists(filename))
    cmd = "grep -v -E '^\s*#\s*include' %s " % filename

    if autoconf_h and os.path.exists(autoconf_h):
        cmd += ' | cpp -include %s' % autoconf_h

    (lines, _) = execute(cmd, echo=True, failok=True)
    # we ignore the exitcode here as we are not interested in failing
    # for #error directives and similar.
    return len(lines)


def find_configurations(filename):
    """
    Collects all configurations that undertaker has created

    To be called right after invoking undertaker's coverage mode.
    """
    ret = list()
    for c in glob.glob(filename + ".config*"):
        if re.match(r".*\.config\d+$", c):
            ret.append(c)
    return sorted(ret)



#
#   golem - analyzes feature dependencies in Linux makefiles
#
# Copyright (C) 2012 Christian Dietrich <christian.dietrich@informatik.uni-erlangen.de>
# Copyright (C) 2012 Reinhard Tartler <tartler@informatik.uni-erlangen.de>
# Copyright (C) 2012 Andreas Ruprecht <rupran@einserver.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from vamos.Config import Config
from vamos.tools import execute, CommandFailed, execute

from tempfile import mkstemp
from glob import glob

import logging
import os
import re
import sys

import vamos

class TreeNotConfigured(RuntimeError):
    """ Indicates that this Linux tree is not configured yet """
    pass


class NotALinuxTree(RuntimeError):
    """ Indicates we are not in a Linux tree """
    pass


def find_autoconf():
    """ returns the path to the autoconf.h file in this linux tree
    """

    if vamos.golem.autoconf_h:
        return vamos.golem.autoconf_h

    (autoconf, _) = execute("find include -name autoconf.h", failok=False)
    autoconf = filter(lambda x: len(x) > 0, autoconf)
    if len(autoconf) != 1:
        logging.error("Found %d autoconf.h files (%s)",
                      len(autoconf), ", ".join(autoconf))
        raise RuntimeError("Not exactly one autoconf.h was found")
    vamos.golem.autoconf_h = autoconf[0]
    return vamos.golem.autoconf_h


def apply_configuration(arch=None, subarch=None, filename=None):
    """
    Applies the current configuration

    This method updates 'include/config/auto.conf' and
    'include/generated/autoconf.h' to match the current configuration.
    Expects a complete configuration in '.config'. If it does not exist,
    the standard configuration 'allyesconfig' is configured.

    If not applying for the default architecture 'x86', the optional
    parameters "arch" (and possibly "subarch") need to be specified.

    If an optional filename is passed, then architecture and subarch are
    guessed using the guess_arch_from_filename() function. Overriding
    either arch or subarch remains possible by explicitly setting arch
    or subarch.
    """

    if filename:
        (guessed_arch, guessed_subarch) = guess_arch_from_filename(filename)
        if not arch:
            arch = guessed_arch
        if not subarch:
            subarch = guessed_subarch

    if not arch:
        logging.warning("No architecture selected. Defaulting to x86")
        arch = 'x86'

    if not os.path.exists(".config"):
        call_linux_makefile("allyesconfig", arch=arch, subarch=subarch)

    # this catches unset defaults. Since boolean and tristate have
    # implicit defaults, this can effectively only happen for integer
    # and hex items. Both are fine with a setting of '0'
    execute('sed -i s,=$,=0,g .config', failok=False)
    try:
        call_linux_makefile('silentoldconfig', arch=arch, subarch=subarch,
                            failok=False)
    except CommandFailed as e:
        if e.returncode == 2:
            raise TreeNotConfigured("target 'silentoldconfig' failed")
        else:
            raise

def guess_source_for_target(target):
    """
    for the given target, try to determine its source file.

    return None if no source file could be found
    """
    for suffix in ('.c', '.S', '.s', '.l', '.y', '.ppm'):
        sourcefile = target[:-2] + suffix
        if os.path.exists(sourcefile):
            return sourcefile
    return None


def files_for_current_configuration(arch=None, subarch=None, how=False):
    """
    to be run in a Linux source tree.

    Returns a list of files that are compiled with the current
    configuration.

    The parameter 'how', which defaults to False, decides on the mode of
    operation. If it is set to False, this function guesses for each
    found file the corresponding source file. A warning is issued if the
    source file could not be guessed and the file is ignored. If it is
    set to True, the function returns a list of object files and
    additionally indicates in what way (i.e., module or statically
    built-in) the object file is compiled.

    NB: If not working for the default architecture 'x86', the optional
    parameters "arch" (and possibly "subarch") need to be specified!
    """

    # locate directory for supplemental makefiles
    scriptsdir = find_scripts_basedir()
    assert os.path.exists(os.path.join(scriptsdir, 'Makefile.list_recursion'))

    apply_configuration(arch=arch, subarch=subarch)

    make_args="-f %(basedir)s/Makefile.list UNDERTAKER_SCRIPTS=%(basedir)s" % \
        { 'basedir' : scriptsdir }

    (output, _) = call_linux_makefile('list',
                                      arch=arch,
                                      subarch=subarch,
                                      failok=False,
                                      extra_variables=make_args)

    files = set()
    for line in output:
        # these lines indicate error and warning messages
        if '***' in line: continue
        if len(line) == 0:
            continue
        try:
            if (how):
                files.add(line)
            else:
                objfile = line.split()[0]
                # try to guess the source filename
                sourcefile = guess_source_for_target(objfile)
                if not sourcefile:
                    logging.debug("Failed to guess source file for %s",
                                  objfile)
                else:
                    files.add(sourcefile)
        except IndexError:
            raise RuntimeError("Failed to parse line '%s'" % line)

    return files


def file_in_current_configuration(filename, arch=None, subarch=None):
    """
    to be run in a Linux source tree.

    Returns the mode (as a string) the file is compiled in the current
    configuration:

        "y" - statically compiled
        "m" - compiled as module
        "n" - not compiled into the kernel

    NB: this function expects the current configuration to be already
        applied with the apply_configuration() function

    """

    # locate directory for supplemental makefiles
    scriptsdir = find_scripts_basedir()
    assert(os.path.exists(os.path.join(scriptsdir, 'Makefile.list_recursion')))

    # normalize filename
    filename = os.path.normpath(filename)

    basename = filename.rsplit(".", 1)[0]
    logging.debug("checking file %s", basename)

    make_args = "-f %(basedir)s/Makefile.list UNDERTAKER_SCRIPTS=%(basedir)s compiled='%(filename)s'" % \
        { 'basedir' : scriptsdir,
          'filename': filename.replace("'", "\'")}

    (make_result, _) = call_linux_makefile('list',
                                           filename=filename,
                                           arch=arch,
                                           subarch=subarch,
                                           failok=False,
                                           extra_variables=make_args)

    for line in make_result:
        # these lines indicate error and warning messages
        if '***' in line: continue
        if line == '': continue
        if line.startswith(basename):
            try:
                return os.path.normpath(line.split(" ")[1])
            except IndexError:
                logging.error("Failed to parse line: '%s'",  line)
                return "n"

    return "n"


def determine_buildsystem_variables(arch=None):
    """
    returns a list of kconfig variables that are mentioned in Linux Makefiles
    """
    if arch:
        cmd = r"find . \( -name Kbuild -o -name Makefile \) " + \
              r"\( ! -path './arch/*' -o -path './arch/%(arch)s/*' \) " + \
              r"-exec sed -n '/CONFIG_/p' {} \+"
        cmd = cmd % {'arch': arch}
    else:
        cmd = r"find . \( -name Kbuild -o -name Makefile \) -exec sed -n '/CONFIG_/p' {} \+"
    find_result = execute(cmd, failok=False)

    ret = set()
    # line might look like this:
    # obj-$(CONFIG_MODULES)           += microblaze_ksyms.o module.o
    for line in find_result[0]:
        for m in re.finditer(r'CONFIG_(?P<feature>[A-Za-z0-9_]+)', line):
            config_variable = m.group('feature')
            ret.add(config_variable)
    return ret


def determine_buildsystem_variables_in_directory(directory):
    filenames = []
    kbuild = os.path.join(directory, "Kbuild")
    makefile = os.path.join(directory, "Makefile")

    if os.path.exists(kbuild):
        filenames.append(kbuild)

    if os.path.exists(makefile):
        filenames.append(makefile)

    if os.path.exists(kbuild + ".platforms"):
        filenames.append(kbuild + ".platforms")
        filenames += glob(directory + "/*/Platform")

    ret = set()

    for filename in filenames:
        with open(filename) as fd:
            for line in fd:
                for m in re.finditer(r'CONFIG_(?P<feature>[A-Za-z0-9_]+)', line):
                    config_variable = m.group('feature')
                    ret.add("CONFIG_" + config_variable)

    return ret


def guess_subarch_from_arch(arch):
    """
    For the given architecture, try to guess the best matching subarchitecture.

    This choice is not so obvious for architectures like 'x86', where valid
    values include 'x86_64' for 64bit CPUs and 'i386' for 32bit CPUs.
    """

    subarch = arch
    if arch == 'x86' or arch == 'um':
        if vamos.prefer_32bit:
            subarch = 'i386'
        else:
            subarch = 'x86_64'
    else:
        assert(arch==subarch)

    return subarch


def guess_arch_from_filename(filename):
    """
    Guesses the 'best' architecture for the given filename.

    If the file is in the Linux HAL (e.g., 'arch/arm/init.c', then the
    architecture is deduced from the filename

    Defaults to 'vamos.default_architecture' (default: 'x86')

    returns a tuple (arch, subarch) with the architecture identifier (as
    in the subdirectory part) and the preferred "subarchitecture". The
    latter is used e.g. to disable CONFIG_64BIT on 'make allnoconfig'
    when compiling on a 64bit host (default behavior), unless
    vamos.prefer_32bit is set to False.
    """

    m = re.search("^arch/([^/]*)/", os.path.normpath(filename))
    if m:
        arch = m.group(1)
    else:
        arch = vamos.default_architecture

    subarch = guess_subarch_from_arch(arch)

    forced_64bit = False
    forced_32bit = False

    if arch =='x86' and '.config' in filename:
        c = Config(filename)
        if c.valueOf('CONFIG_X86_64') == 'y' or c.valueOf('CONFIG_64BIT') == 'y':
            subarch = 'x86_64'
            logging.debug("Config %s forces subarch x86_64", filename)
            forced_64bit = True

        if c.valueOf('CONFIG_X86_64') == 'n' or c.valueOf('CONFIG_64BIT') == 'n' \
                or c.valueOf('CONFIG_X86_32') == 'y':
            subarch = 'i386'
            logging.debug("Config %s forces subarch i386", filename)
            forced_32bit = True

    if forced_32bit and forced_64bit:
        logging.error("%s is inconsistent: cannot enable 64bit and 32bit together (using %s)",
                      filename, subarch)

    return (arch, subarch)


def call_linux_makefile(target, extra_env="", extra_variables="",
                        filename=None, arch=None, subarch=None,
                        failok=True, dryrun=False):
    # pylint: disable=R0912
    """
    Invokes 'make' in a Linux Buildtree.

    This utility function hides details how to set make and environment
    variables that influence kbuild. An important variable is 'ARCH'
    (and possibly later 'SUBARCH'), which can be via the corresponding
    variables.

    If a target points to an existing file (or the optional target
    filename is given)the environment variable for ARCH is derived
    according to the follwing rules:

      - if the file is inside an "arch/$ARCHNAME/", use $ARCHNAME
      - if the "arch" variable is set, use that
      - by default use 'default_arch'
      - if the arch is set to 'x86', set ARCH to 'i386', unless the
        "prefer_64bit" parameter is set to 'True'

    If dryrun is True, then the command line is returned instead of the
    command's execution output. This is mainly useful for testing.

    returns a tuple with
     1. the command's standard output as list of lines
     2. the exitcode
    """

    cmd = "make"
    njobs = int(os.sysconf('SC_NPROCESSORS_ONLN') * 1.20 + 0.5)

    if extra_env and "ARCH=" in extra_env or extra_variables and "ARCH=" in extra_variables:
        logging.debug("Detected manual (SUB)ARCH override in extra arguments '(%s, %s)'",
                      extra_env, extra_variables)
    else:
        if os.path.exists(target):
            filename = target
        if filename:
            (guessed_arch, guessed_subarch) = guess_arch_from_filename(filename)

            if not arch:
                arch = guessed_arch

            if not subarch:
                subarch = guessed_subarch

    if not arch:
        (arch, subarch) = guess_arch_from_filename('Makefile')

    if not subarch:
        subarch = arch

    if arch == 'x86':
        # x86 is special
        if subarch == 'i386' or vamos.prefer_32bit:
            variant = 'i386'
        else:
            variant = 'x86_64'
        extra_env += " ARCH=%s" % variant
    else:
        extra_env += " ARCH=%s" % arch
    extra_env += " SUBARCH=%s" % subarch

    if not 'KERNELVERSION=' in extra_variables:
        if not vamos.kernelversion:
            (output, rc) = execute("git describe", failok=True)
            if rc == 0:
                vamos.kernelversion = output[-1]
        if vamos.kernelversion:
            extra_env += ' KERNELVERSION="%s"' % vamos.kernelversion

    cmd = "env %(extra_env)s make -j%(njobs)s %(target)s %(extra_variables)s " % \
        { 'njobs': njobs,
          'extra_env': extra_env,
          'target': target,
          'extra_variables': extra_variables }

    if dryrun:
        return (cmd, 0)
    else:
        return execute(cmd, failok=failok)


def get_linux_version():
    """
    Checks that the current working directory is actually a Linux Tree

    Uses a custom Makefile to retrieve the current kernel version. If we
    are in a git tree, additionally compare the git version with the
    version stated in the Makefile for plausibility.

    Raises a 'NotALinuxTree' exception if the version could not be retrieved.
    """

    scriptsdir = find_scripts_basedir()

    if not os.path.exists('Makefile'):
        raise NotALinuxTree("No 'Makefile' found")

    if os.path.isdir('.git'):
        cmd = "git describe"
        (output, ret) = execute(cmd)
        git_version = output[0]
        if (ret > 0):
            git_version = ""
            logging.info("Execution of '%s' command failed, analyzing the Makefile instead", cmd)

        # 'standard' Linux repository descriptions start with v
        if git_version.startswith(("v3.", "v2.6")):
            return git_version

    extra_vars = "-f %(basedir)s/Makefile.version UNDERTAKER_SCRIPTS=%(basedir)s" % \
        { 'basedir' : scriptsdir }

    (output, ret) = call_linux_makefile('', extra_variables=extra_vars)
    if ret > 0:
        raise NotALinuxTree("The call to Makefile.version failed")

    version = output[-1] # use last line, if not configured we get additional warning messages
    if not version.startswith(("3.", "2.6")):
        raise NotALinuxTree("Only Linux versions 2.6 and 3.x are supported")
    else:
        return version


def find_scripts_basedir():
    executable = os.path.realpath(sys.argv[0])
    base_dir   = os.path.dirname(executable)
    for d in [ '../lib', '../scripts']:
        f = os.path.join(base_dir, d, 'Makefile.list')
        if os.path.exists(f):
            return os.path.realpath(os.path.join(base_dir, d))
    raise RuntimeError("Failed to locate Makefile.list")


def files_for_selected_features(features, arch, subarch):
    """
    to be run in a Linux source tree.

    The parameter features represents a (partial) Kconfig selection in
    the form of a dict from feature -> value e.g: {'CONFIG_X86': 'y',
    'CONFIG_BARFOO': 'm'}.

    @return a tuple of (files, dirs). The first member 'files' is a list
    of files that are compiled with the selected features
    configuration. The second member 'dirs' is a list if directories
    that are traversed thereby.
    """

    # locate directory for supplemental makefiles
    scriptsdir = find_scripts_basedir()
    assert(os.path.exists(os.path.join(scriptsdir, 'Makefile.list_recursion')))

    (fd, tempfile) = mkstemp()
    with os.fdopen(fd, "w+") as fd:
        logging.info("dumping partial configuration with %d items to %s", len(features.items()), tempfile)
        for (key, value) in features.items():
            fd.write("%s=%s\n" % (key, value))
            logging.info("%s=%s", key, value)

    make_args= "-f %(basedir)s/Makefile.list UNDERTAKER_SCRIPTS=%(basedir)s" % \
        { 'basedir' : scriptsdir }
    make_args += " print_dirs=y print_files=y auto_conf=%(tempfile)s" % \
           {'tempfile': tempfile }

    try:
        (make_result, _) = call_linux_makefile('list',
                                               arch=arch,
                                               subarch=subarch,
                                               failok=False,
                                               extra_variables=make_args)
    except:
        os.unlink(tempfile)
        raise

    files = set()
    dirs = set()
    for line in make_result:
        # these lines indicate error and warning messages
        if '***' in line: continue
        l = line.split()
        try:
            if line.endswith("/"):
                dirs.add(os.path.normpath(line))
            else:
                files.add(os.path.normpath(l[0]))
        except IndexError:
            raise RuntimeError("Failed to parse line '%s'" % line)

    os.unlink(tempfile)

    return (files, dirs)



#
#   rsf2model - extracts presence implications from kconfig dumps
#
# Copyright (C) 2011 Christian Dietrich <christian.dietrich@informatik.uni-erlangen.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import shlex
from vamos.rsf2model.BoolRewriter import BoolRewriter
from vamos.rsf2model.helper import BoolParserException
from vamos.rsf2model import tools

class RsfReader:
    def __init__(self, fd):
        """Read rsf file and store all relations within in database"""
        keys = ["Item", "HasPrompts", "Default", "ItemSelects", "Depends",
                "Choice", "ChoiceItem"]

        self.database = {}
        for key in keys:
            self.database[key] = []

        for line in fd.readlines():
            try:
                row = shlex.split(line)
            except ValueError:
                print "Couldn't parse %s" % line
                continue
            if len(row) > 1 and row[0] in keys:
                self.database[row[0]].append(row[1:])

    @staticmethod
    def symbol(name):
        if " " in name:
            raise OptionInvalid()

        return "CONFIG_%s" % name

    @staticmethod
    def symbol_module(name):
        if " " in name:
            raise OptionInvalid()

        return "CONFIG_%s_MODULE" % name

    @tools.memoized
    def options(self):
        """Returns all configuration options"""

        # Collect all options
        tristate = {}
        omnipresent = {}
        options = set()
        for item in self.database["Item"]:
            if not item[1].lower() in ["boolean", "tristate", "integer", "string", "hex"]:
                continue
            if len(item) < 2:
                continue
            options.add(item[0])
            tristate[item[0]] = item[1].lower() == "tristate"
            omnipresent[item[0]] = False

        # Map them to options
        result = {}
        for option_name in options:
            result[option_name] = Option(self, option_name, tristate[option_name], omnipresent[option_name])

        for item in self.database["Choice"]:
            if len(item) < 2: continue
            tristate = item[2] == "tristate"
            required = item[1] == "required"
            result[item[0]] = Choice(self, item[0], tristate = tristate, required = required)
            if tristate:
                result[item[0] + "_META"] = ChoiceMeta(result[item[0]])

        return result


    @tools.memoized
    def collect(self, key, col = 0, multival = False):
        """Collect all database keys and put them by the n'th column in a dict"""

        result = {}
        for item in self.database[key]:
            if len(item) < col: continue
            if multival:
                if item[col] in result:
                    result[item[col]].append(item[0:col] + item[col+1:])
                else:
                    result[item[col]] = [item[0:col] + item[col+1:]]
            else:
                result[item[col]] = item[0:col] + item[col+1:]
        return result

    @tools.memoized
    def depends(self):
        deps = self.collect("Depends", 0, True)
        for k, v in deps.items():
            v = map(lambda x: x[0], v)
            if len(v) > 1:
                deps[k] = ["(" + ") || (".join(v) + ")"]
            else:
                deps[k] = v
        return deps

    def is_bool_tristate(self, symbol):
        """Returns true if symbol is boolean or tristate, otherwise false is returned."""

        return self.get_type(symbol) in ["boolean", "tristate"]

    def get_type(self, symbol):
        """Get data type of symbol. Returns 'None' if item is not found"""

        for (item, kconfig_type) in self.database["Item"]:
            if item == symbol:
                return kconfig_type
        return None

class OptionInvalid(Exception):
    pass

class OptionNotTristate(Exception):
    pass

class Option (tools.Repr):
    def __init__(self, rsf, name, tristate = False, omnipresent = False):
        tools.Repr.__init__(self)
        self.rsf = rsf
        self.name = name
        self._tristate = tristate
        self._omnipresent = omnipresent

    def omnipresent(self):
        return self._omnipresent

    def set_omnipresent(self, value):
        self._omnipresent = value

    def tristate(self):
        return self._tristate

    def symbol(self):
        if " " in self.name:
            raise OptionInvalid()

        return "CONFIG_%s" % self.name

    def symbol_module(self):
        if " " in self.name:
            raise OptionInvalid()

        if not self._tristate:
            raise OptionNotTristate()
        return "CONFIG_%s_MODULE" % self.name

    def prompts(self):
        prompts = self.rsf.collect("HasPrompts")
        return int(prompts.get(self.name, ["-1"])[0])

    def get_type(self):
        return self.rsf.get_type(self.name)

    def hex(self):
        return self.get_type() == 'hex'

    def string(self):
        return self.get_type() == 'string'

    def dependency(self, eval_to_module = True):
        depends = self.rsf.depends()
        if not self.name in depends or len(depends[self.name]) == 0:
            return None

        depends = depends[self.name][0]

        # Rewrite the dependency
        try:
            return str(BoolRewriter(self.rsf, depends, eval_to_module = eval_to_module).rewrite())
        except BoolParserException:
            return ""

    def has_depends(self):
        depends = self.rsf.depends()

        if not self.name in depends or len(depends[self.name]) == 0:
            return False
        return True

    def __unicode__(self):
        return u"<Option %s, tri: %s>" % (self.name, str(self.tristate()))

class Choice(Option):
    def __init__(self, rsf, name, tristate, required):
        Option.__init__(self, rsf, name, tristate, False)

        ## In case of no dependencies::
        # Boolean tristates are always on, when choice is tristate, we
        # have a ChoiceMeta option in the symbol dict
        self.set_omnipresent(not tristate and not self.has_depends())

        self._required = required

    def required(self):
        return self._required

    def insert_forward_references(self):
        """Insert dependencies SYMBOL -> CHOICE"""
        items = self.rsf.collect("ChoiceItem", 1, True)
        deps = {}
        own_items = []
        deps[self.rsf.symbol(self.name)] = []

        if self.tristate():
            deps[self.rsf.symbol_module(self.name)] = []
            # If we are tristate there is an ALWAYS_ON item
            # CHOICE_%d_META.
            # This is also a dependency of both items.
            deps[self.symbol()].append("%s_META" % self.symbol())
            deps[self.symbol_module()].append("%s_META" % self.symbol())


        if self.name in items:
            for [symbol] in items[self.name]:
                if symbol in self.rsf.options():
                    own_items.append(symbol)
                    deps[self.rsf.symbol(symbol)] = [self.rsf.symbol(self.name)]
                    if self.tristate():
                        # If the choice is tristate the CHOICE_MODULE
                        # implies, that no option from the choice is
                        # selected as static unit
                        deps[self.rsf.symbol_module(self.name)].append("!" + self.rsf.symbol(symbol))
                        opt = self.rsf.options()[symbol]
                        if opt.tristate():
                            deps[self.rsf.symbol_module(symbol)] = [self.rsf.symbol_module(self.name)]


        or_clause = []
        and_clause_count = len(own_items)
        if not self._required:
            # If we have an optional choice, also no item can be
            # selected. So we add an and clause where all items are negated.
            and_clause_count += 1
        for x in range(0, and_clause_count):
            and_clause = []
            for y in range(0, len(own_items)):
                if x == y:
                    and_clause.append(self.rsf.symbol(own_items[y]))
                else:
                    and_clause.append("!" + self.rsf.symbol(own_items[y]))
            or_clause.append(" && ".join(and_clause))
        deps[self.rsf.symbol(self.name)].extend([ "((" + ") || (".join(or_clause) + "))"])

        return deps

class ChoiceMeta(Option):
    def __init__(self, choice):
        self.choice = choice
        Option.__init__(self, choice.rsf, choice.name + "_META", tristate=False, omnipresent=False)
        self.set_omnipresent(not choice.has_depends())

    def dependency(self, eval_to_module = True):
        return "((%s && !%s_MODULE) || (!%s && %s_MODULE))" % \
               tuple([self.choice.name] * 4)




#
#   utility classes for working in source trees
#
# Copyright (C) 2011 Christian Dietrich <christian.dietrich@informatik.uni-erlangen.de>
# Copyright (C) 2011 Reinhard Tartler <tartler@informatik.uni-erlangen.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import re
import logging


class MessageContainer(set):
    def add_message(self, configuration, message):
        for msg in self:
            if msg == message:
                msg.message_is_in_configuration(configuration)
                return
        # Message isn't in configuration already, so just add it
        self.add(message)


class SparseMessage:
    @staticmethod
    def preprocess_messages(lines):
        """Remove the linebreaks introduced by gcc and sparse depending on the line prefix"""
        last_prefix = None
        result = []
        # Remove empty lines
        lines = filter(lambda x: len(x) > 0, lines)
        # Strip coloumn numbers
        lines = map(lambda x: re.sub(r'(:\d+):\d+: ', r'\1: ', x), lines)
        for line in lines:
            m = re.match('^([^ \t]+)([ \t]+)(.*)', line)
            if not m:
                logging.error("Sparse: Failed to parse '%s'", line)
                continue

            if 'originally declared here' in line:
                continue

            if '***' in line:
                continue

            if last_prefix != None and m.group(1) == last_prefix:
                result[-1] += " | " + m.group(3)
                continue
            last_prefix = m.group(1)
            result.append(line)

        return result


    def __init__(self, configuration, line):
        self.configuration = configuration
        self.bare_message = line
        self.in_configurations = set([configuration])
        self.message = ""
        self.location = ""
        self.parse()
        self.is_error = False


    def message_is_in_configuration(self, configuration):
        """Tell this message that it is also in another configuration"""
        self.in_configurations.add(configuration)


    def parse(self):
        message = self.bare_message.split(" ", 2)
        if message[1] == "error:":
            self.is_error = True
        elif message[1] == "warning:":
            self.is_error = False
        else:
            raise RuntimeError("Sparse: Failed to determine criticality of '%s'" % self.bare_message)

        self.location = message[0]
        self.message = message[2]


    def __repr__(self):
        return "<CompileMessage: error: " + str(self.is_error) + \
            " loc: " + self.location + " msg: " + self.message + ">"

    def __hash__(self):
        return hash(self.bare_message)


    def __eq__(self, other):
        if not isinstance(other, SparseMessage):
            return False
        return self.bare_message == other.bare_message


    def get_message(self):
        return self.bare_message


    def get_report(self):
        return self.bare_message + "\n  in %d configs. e.g. in %s" \
               % ( len(self.in_configurations), str(self.configuration))


class GccMessage(SparseMessage):
    @staticmethod
    def preprocess_messages(messages):
        # Remove [-W<flag>]$ messages
        expr = re.compile("\s*\[(-W[^[]+|enabled by default)\]$")
        messages = map(lambda x: re.sub(expr, "", x), messages)
        messages = SparseMessage.preprocess_messages(messages)
        messages = filter(lambda x: re.match(".*:[0-9]+: (warning|error):", x), messages)
        messages = map(lambda x: re.sub(r'(:\d+):\d+: ', r'\1: ', x), messages)
        return messages

    def __init__(self, configuration, line):
        SparseMessage.__init__(self, configuration, line)

    def __hash__(self):
        return self.location.__hash__()

    def __eq__(self, other):
        if not isinstance(other, GccMessage):
            return False
        ret = self.bare_message == other.bare_message

        if self.configuration == other.configuration \
           and self.location == other.location:
            ret = True

        return ret


class ClangMessage(SparseMessage):
    @staticmethod
    def preprocess_messages(messages):
        # Remove [-W<flag>]$ messages
        expr = re.compile("\s*\[(-W[^[]+|enabled by default)\]$")
        messages = map(lambda x: re.sub(expr, "", x), messages)

        messages = filter(lambda x: re.match(".*:[0-9]+:.*(warning|error):", x), messages)
        return messages


    def __init__(self, configuration, line):
        SparseMessage.__init__(self, configuration, line)


    def parse(self):
        message = self.bare_message.split(" ", 2)
        if "error:" == message[1] or "fatal" == message[1]:
            self.is_error = True
        elif message[1] == "warning:" or message[1] == "note:":
            self.is_error = False
        else:
            raise RuntimeError("Clang: Couldn't parse " + self.bare_message)

        self.location = message[0]
        self.message = message[2]

        self.in_configurations = set([self.configuration])



#!/usr/bin/env python
#
#   rsf2model - extracts presence implications from kconfig dumps
#
# Copyright (C) 2011 Christian Dietrich <christian.dietrich@informatik.uni-erlangen.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import unittest as t
import StringIO
from vamos.rsf2model import RsfReader
from vamos.rsf2model.BoolRewriter import BoolRewriter as BR



class TestRsfReader(t.TestCase):
    def setUp(self):
        rsf = """Item A 'boolean'
Item B tristate
ItemFoo B some_value
Item C boolean
Item S string
Item H hex
CRAP CARASDD"""
        self.rsf = RsfReader.RsfReader(StringIO.StringIO(rsf))

    def test_simple_expr(self):
        def rewrite(a, b, to_module = True):
            self.assertEqual(str(BR(self.rsf, a, eval_to_module = to_module).rewrite()), b)

        rewrite("64BIT", "CONFIG_64BIT")
        rewrite("FOO && 64BIT", "(CONFIG_FOO && CONFIG_64BIT)")
        rewrite("FOO && (64BIT)", "(CONFIG_FOO && CONFIG_64BIT)")


        rewrite("A", "CONFIG_A")
        rewrite("!A", "!CONFIG_A")
        rewrite("!(A && C)", "(!CONFIG_A || !CONFIG_C)")

        rewrite("B", "(CONFIG_B_MODULE || CONFIG_B)")
        rewrite("!B", "!CONFIG_B")
        rewrite("B", "CONFIG_B", False)
        rewrite("!B", "(!CONFIG_B_MODULE && !CONFIG_B)", False)

        rewrite("S", "CONFIG_S")
        rewrite("H", "CONFIG_H")

        rewrite("FOO=y", "CONFIG_FOO")

        rewrite("FOO=n", "(!CONFIG_FOO_MODULE && !CONFIG_FOO)")
        rewrite("FOO=m", "CONFIG_FOO_MODULE")
        rewrite("FOO!=y", "!CONFIG_FOO")
        rewrite("FOO!=n", "(CONFIG_FOO_MODULE || CONFIG_FOO)")
        rewrite("FOO!=m", "!CONFIG_FOO_MODULE")

        rewrite("FOO=BAR",
                "((CONFIG_FOO && CONFIG_BAR) || (CONFIG_FOO_MODULE && CONFIG_BAR_MODULE) "
                + "|| (!CONFIG_FOO && !CONFIG_BAR && !CONFIG_FOO_MODULE && !CONFIG_BAR_MODULE))")

        rewrite("FOO!=BAR", "((CONFIG_FOO && !CONFIG_BAR) "
                + "|| (CONFIG_FOO_MODULE && !CONFIG_BAR_MODULE) "
                + "|| (!CONFIG_FOO && CONFIG_BAR) "
                + "|| (!CONFIG_FOO_MODULE && CONFIG_BAR_MODULE))")

        rewrite("FOO && BAR", "(CONFIG_FOO && CONFIG_BAR)")

        rewrite("FOO && BAR=HURZ",
                "(CONFIG_FOO && "
                + "((CONFIG_BAR && CONFIG_HURZ) || (CONFIG_BAR_MODULE && CONFIG_HURZ_MODULE) || "
                + "(!CONFIG_BAR && !CONFIG_HURZ && !CONFIG_BAR_MODULE && !CONFIG_HURZ_MODULE)))")

        rewrite("FOO && BAR=FOO",
                "(CONFIG_FOO && "
                + "((CONFIG_BAR && CONFIG_FOO) || (CONFIG_BAR_MODULE && CONFIG_FOO_MODULE) || "
                + "(!CONFIG_BAR && !CONFIG_FOO && !CONFIG_BAR_MODULE && !CONFIG_FOO_MODULE)))")

        rewrite("B43 && (HW_RANDOM || HW_RANDOM=B43)",
                "(CONFIG_B43 && "
                + "(CONFIG_HW_RANDOM || "
                + "((CONFIG_HW_RANDOM && CONFIG_B43) || "
                + "(CONFIG_HW_RANDOM_MODULE && CONFIG_B43_MODULE) || "
                + "(!CONFIG_HW_RANDOM && !CONFIG_B43 && "
                + "!CONFIG_HW_RANDOM_MODULE && !CONFIG_B43_MODULE))))")

        self.rsf = RsfReader.RsfReader(open("validation/equals-module.fm"))

        rewrite("SSB_POSSIBLE && SSB && (PCI || PCI=SSB)",
                "(CONFIG_SSB_POSSIBLE && (CONFIG_SSB_MODULE || CONFIG_SSB) && "
                + "((CONFIG_PCI_MODULE || CONFIG_PCI) || ((CONFIG_PCI && CONFIG_SSB) || "
                + "(CONFIG_PCI_MODULE && CONFIG_SSB_MODULE) || "
                + "(!CONFIG_PCI && !CONFIG_SSB && !CONFIG_PCI_MODULE && !CONFIG_SSB_MODULE))))")

        rewrite("THINKPAD_ACPI && SND && (SND=y || THINKPAD_ACPI=SND)",
                "(CONFIG_THINKPAD_ACPI && CONFIG_SND && "
                + "(CONFIG_SND || ((CONFIG_THINKPAD_ACPI && CONFIG_SND) || "
                + "(CONFIG_THINKPAD_ACPI_MODULE && CONFIG_SND_MODULE) || "
                + "(!CONFIG_THINKPAD_ACPI && !CONFIG_SND && !CONFIG_THINKPAD_ACPI_MODULE && !CONFIG_SND_MODULE))))")


if __name__ == '__main__':
    t.main()



#
#   rsf2model - extracts presence implications from kconfig dumps
#
# Copyright (C) 2012 Manuel Zerpies <manuel.f.zerpies@ww.stud.uni-erlangen.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

def tree_change(func, tree):
    """
    Calls func on every subtree and replaces it with the result if
    not None, otherwise we need to go deeper
    """

    if not type(tree) in (list, tuple) or len(tree) < 1:
        return tree
    a = func(tree)
    if a != None:
        return a
    i = 1

    while i < len(tree):
        b = tree_change(func, tree[i])
        if b == []:
            del tree[i]
        else:
            tree[i] = b
            i+= 1
    return tree


class BoolParserException(RuntimeError):
    def __init__(self, value):
        RuntimeError.__init__(self, value)

    def __unicode__(self):
        return "ERROR: " + self.value
    __repr__ = __unicode__

class BoolRewriterException(RuntimeError):
    def __init__(self, value):
        RuntimeError.__init__(self)
        self.value = value

    def __unicode__(self):
        return "ERROR: " + self.value
    __repr__ = __unicode__



#
#   golem - analyzes feature dependencies in Linux makefiles
#
# Copyright (C) 2012 Christian Dietrich <christian.dietrich@informatik.uni-erlangen.de>
# Copyright (C) 2012 Reinhard Tartler <tartler@informatik.uni-erlangen.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import logging
import os

from vamos.golem.kbuild import determine_buildsystem_variables_in_directory, \
    apply_configuration, guess_source_for_target, call_linux_makefile

from vamos.selection import Selection
from vamos.golem.FileSet import FileSet, FileSetCache

def objects_in_dir(directory):
    """
    Returns object files and subdirectories in a specific directory

    @return tuple(set(object_files), set(subdirectories))
    """

    ret = (set(), set())

    for name in os.listdir(directory):
        name = os.path.join(directory, name)
        # is a directory
        if os.path.isdir(name):
            ret[1].add(name)
        elif name.endswith(".c"):
            name = name[:-len(".c")] + ".o"
            ret[0].add(name)
    return ret


class FileSymbolInferencer:
    def __init__(self, model, arch, subarch=None, subdir=""):
        """
        subdir is used to limit the subdirectories, on which the
        inference algorithm operates.

        """

        self.arch = arch
        self.subarch = subarch
        self.model = model

        # declaration only
        self.empty_selection = Selection()
        self.empty_fileset = None

        self.file_selections = {}

        # Directories that caused a recursion loop
        self.visited_dirs = set()
        self.directory_prefix = subdir

        self.file_set_cache = FileSetCache(arch=arch, subarch=subarch)


    def calculate(self):
        call_linux_makefile('allnoconfig', arch=self.arch, subarch=self.subarch)
        apply_configuration(arch=self.arch, subarch=self.subarch)

        self.empty_fileset = FileSet(self.empty_selection, arch=self.arch,
                                     subarch=self.subarch)

        for filename in self.empty_fileset.files:
            self.file_selections[filename] = None

        directories = self.empty_fileset.dirs
        directories.add("arch/%s" % self.arch)

        for directory in directories:
            symbol_slice = determine_buildsystem_variables_in_directory(directory)
            self.__calculate(directory, symbol_slice,
                             self.empty_selection,
                             self.empty_fileset)

        for objfile in self.file_selections:
            sourcefile = guess_source_for_target(objfile)
            if not sourcefile:
                logging.warning("Failed to guess source file for %s", objfile)
                sourcefile = objfile # makes these cases easier to identify
            if self.file_selections[objfile]:
                print "FILE_%s \"%s\"" % (sourcefile, self.file_selections[objfile])
            else:
                print "FILE_%s" % sourcefile

    def __subdirectory_is_worth_working_on(self, directory):
        """
        Determine if it is worth to work on a specific directory,
        or if all object files and all directories within are already visited
        """

        (objects, dirs) = objects_in_dir(directory)

        if not directory.startswith(self.directory_prefix):
            logging.info("Skipping %s, not in scope", directory)
            return False

        # Are all objects visited?
        for obj in objects:
            if not obj in self.file_selections:
                return True

        # Are all directories visited
        for dirname in dirs:
            if not dirname in self.visited_dirs:
                return True

        logging.info("Skipping %s, already processed", directory)
        return False


    def check_subdir(self, subdir, symbols):
        """
        @return a set of new symbols found in subdir
        """

        new_symbols = set()
        new_symbols.update(determine_buildsystem_variables_in_directory(subdir))
        new_symbols.update(symbols)

        # For tristate features also add the _MODULE
        # symbol
        to_be_removed = set()
        to_be_added = set()
        for symbol in new_symbols:
            # Normalize feartures
            if symbol.endswith("_MODULE"):
                symbol = symbol[:-len("_MODULE")]

            # Add _MODULE symbols
            if (symbol + "_MODULE") in self.model:
                to_be_added.add(symbol + "_MODULE")

            # Remove missing items in the first place
            if not symbol in self.model:
                to_be_removed.add(symbol)
                to_be_removed.add(symbol + "_MODULE")
        new_symbols = (new_symbols - to_be_removed).union(to_be_added)

        return new_symbols

    def __copy_selections(self, directory, base_symbols, current_selection, current_fileset):
        """
        @return a list of tulples: (subdir, symbols, selection, fileset)
        """
        file_selections = {}
        recursive_selections = []
        new_visited_dirs = self.visited_dirs

        for symbol in base_symbols:

            if symbol in current_selection.symbols:
                continue

            if not symbol in self.model:
                continue

            # Copy Selection
            new_selection = Selection(current_selection)
            new_selection.push_down()
            new_selection.add_alternative(symbol)
            new_fileset = self.file_set_cache.get_fileset(new_selection)

            ((f_added, _), (d_added, _)) = new_fileset.compare_to_base(current_fileset)

            for filename in f_added: # f_added
                if not filename in file_selections:
                    file_selections[filename] = Selection(new_selection)
                else:
                    file_selections[filename].add_alternative(symbol)

            # Is there a directory we haven't done recursion on yet,
            # so lets do a recursion step here.
            # We only consider subdirectories of the current directory
            new_subdirectories = d_added - self.visited_dirs
            new_subdirectories = set([dirname for dirname in new_subdirectories
                                      if dirname.startswith(directory) and
                                      dirname.startswith(self.directory_prefix)])
            if len(new_subdirectories) > 0:
                new_visited_dirs = new_visited_dirs.union(set(new_subdirectories))

                # If we want to go down recursive in a directory, we
                # add all symbols which are mentioned within the
                # directories Kbuild/Makefile
                for subdir in new_subdirectories:
                    new_symbols = self.check_subdir(subdir, new_selection.symbols)
                    recursive_selections.append( tuple([subdir,
                                                        new_symbols,
                                                        new_selection,
                                                        new_fileset]))

        self.visited_dirs = new_visited_dirs

        return (recursive_selections, file_selections)


    def __calculate(self, directory, base_symbols, current_selection, current_fileset):
        """
        The heart of the inferencer, here the recursion is done.
        """

        if not self.__subdirectory_is_worth_working_on(directory):
            return

        logging.info("Starting __calculate on %s with %d symbols (%d files so far)",
                     directory, len(base_symbols), len(self.file_selections))

        (recursive_selections, file_selections) =  \
            self.__copy_selections(directory, base_symbols,
                                   current_selection, current_fileset)

        # Collection phase
        for (filename, selection) in file_selections.items():
            if not filename in self.file_selections:
                self.file_selections[filename] = selection
            else:
                if selection.better_than(self.file_selections[filename]):
                    self.file_selections[filename] = selection
                else:
                    # Sometimes Expressions can be merged if they only
                    # differ in one or_expression
                    merged = self.file_selections[filename].merge(selection)
                    if merged:
                        self.file_selections[filename] = merged

        # RECURSION!
        for (subdir, symbols, selection, fileset) in recursive_selections:
            self.__calculate(subdir, symbols, selection, fileset)



#
# Copyright (C) 2011 Reinhard Tartler <tartler@informatik.uni-erlangen.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import unittest2 as t

import vamos

from vamos.tools import execute, CommandFailed
from vamos.golem.kbuild import *


class testTools(t.TestCase):
    def setUp(self):
        pass

    def test_execute(self):
        (output, rc) = execute("false", failok=True)
        self.assertEqual(rc, 1)
        self.assertIs(len(output), 1)
        self.assertIs(len(output[0]), 0)
        with self.assertRaises(CommandFailed):
            execute("false", failok=False)

    def test_guessfilename(self):
        vamos.prefer_32bit = False

        (arch, subarch) = guess_arch_from_filename('arch/x86/init.c')
        self.assertEqual(arch, 'x86')
        self.assertEqual(subarch, 'x86_64')

        vamos.prefer_32bit = False
        (arch, subarch) = guess_arch_from_filename('arch/x86/../x86/init.c')
        self.assertEqual(arch, 'x86')
        self.assertEqual(subarch, 'x86_64')

        (arch, subarch) = guess_arch_from_filename('kernel/init.c')
        self.assertEqual(arch, 'x86')
        self.assertEqual(subarch, 'x86_64')

        (arch, subarch) = guess_arch_from_filename('././kernel/init.c')
        self.assertEqual(arch, 'x86')
        self.assertEqual(subarch, 'x86_64')

        vamos.prefer_32bit = True
        (arch, subarch) = guess_arch_from_filename('arch/x86/init.c')
        self.assertEqual(arch, 'x86')
        self.assertEqual(subarch, 'i386')

        (arch, subarch) = guess_arch_from_filename('arch/sparc/init.c')
        self.assertEqual(arch, 'sparc')
        self.assertEqual(subarch, 'sparc')

        (arch, subarch) = guess_arch_from_filename('arch/h8300/kernel/timer/tpu.c')
        self.assertEqual(arch, 'h8300')
        self.assertEqual(subarch, 'h8300')

        (arch, subarch) = guess_arch_from_filename('kernel/init.c')
        self.assertEqual(arch, 'x86')
        self.assertEqual(subarch, 'i386')

    def test_call_linux_makefile(self):

        vamos.prefer_32bit = True

        (cmd, rc) = call_linux_makefile('', dryrun=True)
        self.assertEqual(rc, 0)
        for needle in ('make', ' ARCH=i386', 'SUBARCH=i386'):
            self.assertIn(needle, cmd)

        (cmd, rc) = call_linux_makefile('allnoconfig', dryrun=True)
        self.assertEqual(rc, 0)
        for needle in ('make', 'allnoconfig', ' ARCH=i386', 'SUBARCH=i386'):
            self.assertIn(needle, cmd)

        (cmd, rc) = call_linux_makefile('silentoldconfig', dryrun=True)
        self.assertEqual(rc, 0)
        for needle in ('make', 'silentoldconfig', 'ARCH=i386', 'SUBARCH=i386'):
            self.assertIn(needle, cmd)

        (cmd, rc) = call_linux_makefile('', filename='./arch/x86/init.o', dryrun=True)
        self.assertEqual(rc, 0)
        for needle in ('make', 'ARCH=i386', 'SUBARCH=i386'):
            self.assertIn(needle, cmd)

        (cmd, rc) = call_linux_makefile('', filename='./arch/um/init.o', dryrun=True)
        self.assertEqual(rc, 0)
        for needle in ('make', ' ARCH=um', 'SUBARCH=i386'):
            self.assertIn(needle, cmd)

        (cmd, rc) = call_linux_makefile('arch/h8300/kernel/timer/tpu.o',
                                        filename='arch/h8300/kernel/timer/tpu.c',
                                        extra_variables="C=2",
                                        dryrun=True)
        self.assertEqual(rc, 0)
        for needle in ('make', ' ARCH=h8300', 'SUBARCH=h8300', 'C=2', 'tpu.o'):
            self.assertIn(needle, cmd)


        vamos.prefer_32bit = False
        (cmd, rc) = call_linux_makefile('', filename='./arch/x86/init.o', dryrun=True)
        self.assertEqual(rc, 0)
        for needle in ('make', ' ARCH=x86_64', 'SUBARCH=x86_64'):
            self.assertIn(needle, cmd)

        (cmd, rc) = call_linux_makefile('', filename='./arch/um/init.o', dryrun=True)
        self.assertEqual(rc, 0)
        for needle in ('make', ' ARCH=um', 'SUBARCH=x86_64'):
            self.assertIn(needle, cmd)

        (cmd, rc) = call_linux_makefile('arch/h8300/kernel/timer/tpu.o',
                                        filename='arch/h8300/kernel/timer/tpu.c',
                                        extra_variables="C=2",
                                        dryrun=True)
        self.assertEqual(rc, 0)
        for needle in ('make', ' ARCH=h8300', 'SUBARCH=h8300', 'C=2', 'tpu.o'):
            self.assertIn(needle, cmd)

if __name__ == '__main__':
    t.main()



#
#   vamos - common auxiliary functionality
#
# Copyright (C) 2011-2012 Christian Dietrich <christian.dietrich@informatik.uni-erlangen.de>
# Copyright (C) 2011-2012 Reinhard Tartler <tartler@informatik.uni-erlangen.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from vamos.rsf2model.RsfReader import RsfReader

import logging
import re

class Model(dict):
    def __init__(self, path, rsf = None):
        dict.__init__(self)
        self.path = path
        self.always_on_items = set()

        with open(path) as fd:
            self.parse(fd)

        if not rsf and path.endswith(".model"):
            rsf = path[:-len(".model")] + ".rsf"

        if rsf:
            try:
                with open(rsf) as f:
                    self.rsf = RsfReader(f)
            except IOError:
                logging.warning ("no rsf file for model %s was found", path)

    def parse(self, fd):
        for line in fd.readlines():
            line = line.strip()
            if line.startswith("UNDERTAKER_SET ALWAYS_ON"):
                line = line.split(" ")[2:]
                always_on_items = [ l.strip(" \t\"") for l in line]
                for item in always_on_items:
                    self[item] = None
                    self.always_on_items.add(item)
                continue
            elif line.startswith("I:") or line.startswith("UNDERTAKER_SET"):
                continue
            line = line.split(" ", 1)
            if len(line) == 1:
                self[line[0]] = None
            elif len(line) == 2:
                self[line[0]] = line[1].strip(" \"\t\n")


    def mentioned_items(self, key):
        """
        @return list of mentioned items of the key's implications
        """
        expr = self.get(key, "")
        if expr == None:
            return []
        expr = re.split("[()&!|><-]", expr)
        expr = [x.strip() for x in expr]
        return [x for x in expr if len(x) > 0]

    def leaf_features(self):
        """
        Leaf features are feature that aren't mentioned anywhere else

        @return list of leaf features in model
        """

        # Dictionary of mentioned items
        features = dict([(key, False) for key in self.keys()
                        if not key.endswith("_MODULE")])
        for key in features.keys():
            items = self.mentioned_items(key)
            items += self.mentioned_items(key + "_MODULE")
            for mentioned in items:
                # Strip _MODULE POSTFIX
                if mentioned.endswith("_MODULE"):
                    mentioned = mentioned[:-len("_MODULE")]

                # A Leaf can't "unleaf" itself. This is important for
                # circular relations like:
                #
                # CONFIG_A -> !CONFIG_A_MODULE
                # CONFIG_A_MODULE -> !CONFIG_A
                if key != mentioned and mentioned in features:
                    features[mentioned] = True
        return sorted([key for key in features if features[key] == False])

    def get_type(self, symbol):
        """
        RuntimeError gets raised if no corresponding rsf is found.
        @return data type of symbol.
        """
        if not self.rsf:
            raise RuntimeError("No corresponding rsf file found.")
        if symbol.startswith("CONFIG_"):
            symbol = symbol[len("CONFIG_"):]
        return self.rsf.get_type(symbol)


    def is_bool_tristate(self, symbol):
        """
        Returns true if symbol is tristate, otherwise false is returned.

        Raises RuntimeError if no corresponding rsf-file is found.
        """

        if not self.rsf:
            raise RuntimeError("No corresponding rsf file found.")

        if symbol.startswith("CONFIG_"):
            symbol = symbol[len("CONFIG_"):]

        return self.rsf.is_bool_tristate(symbol)

    def slice_symbols(self, initial):
        """
        Apply the slicing algorithm to the given set of symbols
        returns a list of interesting symbol
        """
        if type(initial) == list:
            stack = initial
        else:
            stack = [initial]

        visited = set()

        while len(stack) > 0:
            symbol = stack.pop()
            visited.add(symbol)
            for new_symbols in self.mentioned_items(symbol):
                if not new_symbols in visited:
                    stack.append(new_symbols)
        return list(visited)



#!/usr/bin/env python
#
#   tests undertaker to hande #define statements properly
#
# Copyright (C) 2011-2012 Christian Dietrich <christian.dietrich@informatik.uni-erlangen.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import sys
import subprocess
import getopt

VERBOSITY=1
filename = ""

def call_undertaker(filename):
    """ Generates the presence conditions for `filename' in one line """
    cpppc = subprocess.Popen(['undertaker', '-q', '-j', 'cpppc', filename],
                             stdin=subprocess.PIPE,
                             stdout=subprocess.PIPE,
                             stderr=subprocess.PIPE)
    cpppc.stdin.close()
    (stdout_got, stderr_got) = (cpppc.stdout.readlines(), cpppc.stderr.readlines())
    cpppc.wait()

    if 0 != cpppc.returncode or len(stderr_got) > 0:
        print "STDOUT:", stdout_got
        print "STDERR:", stderr_got
        raise RuntimeError("calling undertaker in cpppc mode on '%s' failed" % filename)

    return ' '.join(stdout_got)

def call_limboole(selection, cpppc):
    """ returns True if (selection) && (cpppc) is satisfiable """
    limboole_in = ' && '.join([selection, cpppc]).replace('&&', '&').replace("||", "|")

    if VERBOSITY == 2:
        print limboole_in

    p = subprocess.Popen(['limboole', '-s'],
                                              stdin=subprocess.PIPE,
                                              stdout = subprocess.PIPE,
                                              stderr = subprocess.PIPE)

    p.stdin.write(limboole_in)
    p.stdin.close()
    (stdout_got, stderr_got) = (p.stdout.readlines(), p.stderr.readlines())
    p.wait()

    if 0 != p.returncode or len(stderr_got) > 0:
        print "STDOUT:", stdout_got
        print "STDERR:", stderr_got
        raise RuntimeError("calling limboole on '%s' failed" % limboole_in)

    return not ' UNSATIS' in stdout_got[0]


def get_limboole_version():
    """ calls limboole and returns the version number. """

    try:
        p = subprocess.Popen(['limboole', '--version'],
                             stdout = subprocess.PIPE,
                             stderr = subprocess.PIPE)
    except OSError:
        raise RuntimeError("calling limboole failed")

    (stdout_got, stderr_got) = (p.stdout.readlines(), p.stderr.readlines())
    p.wait()

    if 0 != p.returncode or len(stderr_got) > 0:
        print "STDOUT:", stdout_got
        print "STDERR:", stderr_got
        raise RuntimeError("calling limboole failed")

    return stdout_got[0]


def for_all_combinations(length, fn):
    """ applies `fn' to all combinations of a boolean string of length `length' """
    def recursor(rest):
        if rest == 0:
            raise RuntimeError("Internal Error: cannot permutate empty list")
        if rest == 1:
            return [[ False ], [ True ]]
        else:
            return [ [False] + i for i in recursor(rest-1) ] +\
                [ [True] + i for i in recursor(rest-1) ]

    combinations = recursor(length)
    return [ fn(i) for i in combinations ]

class Main:
    def __init__(self):
        self.vars = set()
        self.blocks = set()
        self.content = ""

    def cpp_results(self):
        """ generates list containing all selectable sets of cpp blocks
            [ set(['B1', 'B5', 'B12']), set(['B2', 'B5']) ]
        """
        def tester(combination):
            """ generating preprocessor result for specific set of Macros
            [ True, False, True ] -> cpp -DA -DC -> set(['B1', 'B3', 'B4', 'B12'])
            """
            flags = ['cpp']
            for i in zip(combination, list(self.vars)):
                if i[0]:
                    flags.append('-D%s' % i[1])
            flags.append('-')
            blob = subprocess.Popen(flags,
                                    stdin=subprocess.PIPE,
                                    stdout=subprocess.PIPE,
                                    stderr=subprocess.PIPE).communicate(self.content)[0]
            return set([ i.strip() for i in blob.split('\n') if i.startswith('B') ])

        return for_all_combinations(len(self.vars), tester)

    def create_combination(self, combination):
        """ assembles a subset of existing blocks using `combination'
            combination -> [ True, True, False, True ]
            self.blocks -> [ 'B0', 'B3', 'B5', 'B12' ]
            -> [ 'B0', 'B3', 'B12' ] , [ 'B5' ]
        """
        needed = []
        notneeded = []
        for i in zip(combination, self.blocks):
            if i[0]:
                needed.append(i[1])
            else:
                notneeded.append(i[1])

        return needed, notneeded

    def check(self):
        cpppc = call_undertaker(filename)
        cppergs = self.cpp_results()
        if VERBOSITY == 2:
            print cppergs

        def printer(combination):
            needed, notneeded = self.create_combination(combination)
            cppresult = set(needed) in cppergs

            # [ 'B0', 'B3', 'B12' ] , [ 'B5' ] -> "B0 && B3 && B12 && !B5"
            blockstring = ' && '.join(needed + [ '!%s' % i for i in notneeded ] )
            undertakerresult = call_limboole(blockstring, cpppc)

            if VERBOSITY > 0 or cppresult != undertakerresult:
                print blockstring.ljust(7*(len(needed) + len(notneeded))),
                print '\t1' if cppresult else '\t0',
                print '\t1' if undertakerresult else '\t0',
                print '\tFINE' if cppresult == undertakerresult else '\tDIFFERENT'

            return cppresult == undertakerresult

        return False in for_all_combinations(len(self.blocks), printer)

    def main(self):
        p = subprocess.Popen(['predator', '-c', filename], stdout=subprocess.PIPE)
        self.content = p.communicate()[0]
        if p.returncode != 0:
            raise RuntimeError("Error while running predator")

        continue_with_next = False

        def extract_block(line):
            if line.startswith('B'):
                self.blocks.add(line.strip())

        def extract_cppflag(line):
            try:
                if line.startswith('#if') or continue_with_next:
                    continue_with_next = False
                    if "\\" in line:
                        continue_with_next = True
                    for token in ["(", ")", "||", "&&", "defined", "==", '\\']:
                        line = line.replace(token, " ")
                    self.vars = self.vars | set(line.split()[1:])
            except:
                pass

        for line in self.content.split('\n'):
            if VERBOSITY == 2:
                print line
            extract_block(line)
            extract_cppflag(line)
        if VERBOSITY == 2:
            print self.vars
            print self.blocks

def usage():
    print "%s [-v level] filename" % sys.argv[0]
    print "%s -h" % sys.argv[0]
    print
    print "-v: Verbosity"
    print "  -v 0: only print differences"
    print "  -v 1: print result for all combinations"
    print "  -v 2: noisy debug output"
    print
    print "Return values:"
    print " 1: Operation Error"
    print " 2: Some checks didn't pass"

if __name__ == '__main__':
    opts, args = getopt.getopt(sys.argv[1:], "hv:")

    try:
        limboole_version = get_limboole_version()
    except RuntimeError:
        print "Could not execute limboole, check that it is available"
        sys.exit(1)

    for opt,arg in opts:
        if opt in ('-v', '--verbose'):
            VERBOSITY=int(arg)
            if VERBOSITY < 0 or VERBOSITY > 2:
                print "invalid verbosity level"
                usage()
                sys.exit(1)
        elif opt in ('-h', '--help'):
            usage()
            sys.exit(0)
        else:
            usage()
            assert False, "unknown option"

    if len(args) != 1:
        print "(only) expecting filename as argument"
        usage()
        sys.exit(1)

    if VERBOSITY > 1:
        print "Using limboole version: " + limboole_version

    filename = args[0]

    m = Main()
    m.main()
    sys.exit(2 if m.check() else 0)



#
#   utility classes for working in source trees
#
# Copyright (C) 2011 Christian Dietrich <christian.dietrich@informatik.uni-erlangen.de>
# Copyright (C) 2011-2012 Reinhard Tartler <tartler@informatik.uni-erlangen.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from vamos.vampyr.Messages import SparseMessage, GccMessage, ClangMessage
from vamos.vampyr.utils import ExpansionError, ExpansionSanityCheckError
from vamos.golem.kbuild import guess_arch_from_filename, call_linux_makefile, \
    apply_configuration, find_autoconf, guess_subarch_from_arch, \
    files_for_current_configuration
from vamos.tools import execute
from vamos.model import Model
from vamos.Config import Config

import logging
import os.path
import re
import shutil

class Configuration:
    def __init__(self, framework, basename, nth):
        self.cppflags = '%s.cppflags%s' % (basename, nth)
        self.source   = '%s.source%s' % (basename, nth)
        self.kconfig  = '%s.config%s' % (basename, nth)
        # Note, undertaker will clean up ".config*", which matches self.config_h
        self.config_h = '%s.config%s.h' % (basename, nth)
        self.framework = framework
        self.basename = basename
        self.write_config_h(self.config_h)

    def write_config_h(self, config_h):
        with open(config_h, 'w') as fd:
            cppflags = self.get_cppflags()
            flags = cppflags.split("-D")
            logging.debug("Generating %s, found %d items",
                          config_h, len(flags))
            for f in flags:
                if f == '': continue # skip empty fields
                try:
                    name, value = f.split('=')
                    fd.write("#define %s %s\n" % (name, value))
                except ValueError:
                    logging.error("%s: Failed to parse flag '%s'", config_h, f)

    def switch_to(self):
        raise NotImplementedError

    def call_sparse(self, on_file):
        raise NotImplementedError

    def get_cppflags(self):
        with open(self.cppflags, 'r') as fd:
            return fd.read().strip()

    def filename(self):
        return self.kconfig

    def __copy__(self):
        raise RuntimeError("Object <%s> is not copyable" % self)

    def __deepcopy__(self):
        raise RuntimeError("Object <%s> is not copyable" % self)

    def get_config_h(self):
        """
        Returns the path to a config.h like configuration file
        """
        return self.config_h


class BareConfiguration(Configuration):

    def __init__(self, framework, basename, nth):
        Configuration.__init__(self, framework, basename, nth)

    def __repr__(self):
        return '"' + self.get_cppflags() + '"'

    def switch_to(self):
        pass

    def __call_compiler(self, compiler, args, on_file):
        cmd = compiler + " " + args
        if compiler in self.framework.options['args']:
            cmd += " " + self.framework.options['args'][compiler]
        cmd += " " + self.get_cppflags()
        cmd += " '" + on_file + "'"
        (out, returncode) = execute(cmd, failok=True)
        if returncode == 127:
            raise RuntimeError(compiler + " not found on this system?")
        else:
            return (out, returncode)


    def call_sparse(self, on_file):
        (messages, statuscode) = self.__call_compiler("sparse", "", on_file)
        if statuscode != 0:
            messages.append(on_file +
                            ":1: error: cannot compile file [ARTIFICIAL, rc=%d]" % statuscode)

        messages = SparseMessage.preprocess_messages(messages)
        messages = map(lambda x: SparseMessage(self, x), messages)
        return messages


    def call_gcc(self, on_file):
        (messages, statuscode) = self.__call_compiler("gcc", "-o/dev/null -c", on_file)
        if statuscode != 0:
            messages.append(on_file +
                            ":1: error: cannot compile file [ARTIFICIAL, rc=%d]" % statuscode)

        messages = GccMessage.preprocess_messages(messages)
        messages = map(lambda x: GccMessage(self, x), messages)
        return messages


    def call_clang(self, on_file):
        (messages, statuscode) = self.__call_compiler("clang", "--analyze", on_file)

        if statuscode != 0:
            messages.append(on_file +
                            ":1: error: cannot compile file [ARTIFICIAL, rc=%d]" % statuscode)

        messages = ClangMessage.preprocess_messages(messages)
        messages = map(lambda x: ClangMessage(self, x), messages)
        return messages

    def expand(self, verify=True):
        pass


class LinuxConfiguration(Configuration):
    """
    This class represents a (partial) Linux configuration.

    The expand method uses Kconfig to fill up the remaining
    variables. The parameter expansion_strategy of the __init__ method
    selects how partial configurations get expanded. In the default
    mode, 'alldefconfig', the strategy is to set the remaining variables
    to their Kconfig defined defaults. With 'allnoconfig', the strategy
    is to enable as few features as possible.

    The attribute 'model' of this class is allocated in the expand()
    method on demand.

    """
    def __init__(self, framework, basename, nth,
                 arch=None, subarch=None,
                 expansion_strategy='alldefconfig'):
        Configuration.__init__(self, framework, basename, nth)

        self.expanded = None
        self.expansion_strategy = expansion_strategy
        self.model = None
        try:
            os.unlink(self.kconfig + '.expanded')
        except OSError:
            pass

        if arch:
            self.arch = arch

            if subarch:
                self.subarch = subarch
            else:
                self.subarch = guess_subarch_from_arch(arch)
        else:
            oldsubarch = subarch
            self.arch, self.subarch = guess_arch_from_filename(self.kconfig)
            if oldsubarch:
                self.subarch = oldsubarch

        self.result_cache = {}

    def get_config_h(self):
        return find_autoconf()

    def __repr__(self):
        return '<LinuxConfiguration "' + self.kconfig + '">'

    def expand(self, verify=False):
        """
        @raises ExpansionError if verify=True and expanded config does
                               not patch all requested items
        """
        logging.debug("Trying to expand configuration " + self.kconfig)

        if not os.path.exists(self.kconfig):
            raise RuntimeError("Partial configuration %s does not exist" % self.kconfig)

        (files, _) = execute("find include -name autoconf.h -print -delete", failok=False)
        if len(files) > 1:
            logging.error("Deleted spurious configuration files: %s", ", ".join(files))

        target = self.expansion_strategy
        extra_var = 'KCONFIG_ALLCONFIG="%s"' % self.kconfig
        call_linux_makefile(target, extra_variables=extra_var,
                            arch=self.arch, subarch=self.subarch,
                            failok=False)
        apply_configuration(arch=self.arch, subarch=self.subarch)

        self.expanded = self.save_expanded('.config')

        if verify:
            if not self.model:
                modelf = "models/%s.model" % self.arch
                self.model = Model(modelf)
                logging.info("Loaded %d items from %s", len(self.model), modelf)

            all_items, violators = self.verify(self.expanded)
            if len(violators) > 0:
                logging.warning("%d/%d mismatched items", len(violators), len(all_items))
                for v in violators:
                    logging.info("violating item: %s", v)
                raise ExpansionError("Config %s failed to expand properly" % self.kconfig)
            else:
                logging.info("All items are set correctly")

    def save_expanded(self, config):
        expanded_config = self.kconfig + '.expanded'
        shutil.copy(config, expanded_config)
        return expanded_config

    def verify(self, expanded_config='.config'):
        """
        verifies that the given expanded configuration satisfies the
        constraints of the given partial configuration.

        @return (all_items, violators)
          all_items: set of all items in partial configuration
          violators: list of items that violate the partial selection
        """

        partial_config = Config(self.kconfig)
        config = Config(expanded_config)
        conflicts = config.getConflicts(partial_config)

        return (partial_config.keys(), conflicts)


    def switch_to(self):
        if not self.expanded:
            logging.debug("Expanding partial configuration %s", self.kconfig)
            self.expand()
            return

        # now replace the old .config with our 'expanded' one
        shutil.copyfile(self.expanded, '.config')

        call_linux_makefile('oldconfig', failok=False,
                            arch=self.arch, subarch=self.subarch)


        (_, returncode) = call_linux_makefile('silentoldconfig', failok=True,
                                              arch=self.arch, subarch=self.subarch)
        if returncode != 0:
            # with the satconfig approach, this always worked
            m = "silentoldconfig failed while switching to config %s (rc: %d)" \
                % (self.expanded, returncode)
            raise ExpansionSanityCheckError(m)


    def __call_make(self, on_file, extra_args):
        on_object = on_file[:-1] + "o"
        try:
            os.unlink(on_object)
        except OSError:
            pass

        (messages, statuscode) = \
            call_linux_makefile(on_object, failok=True, arch=self.arch, subarch=self.subarch,
                                extra_variables=extra_args)

        state = None
        CC = []
        CHECK = []

        while len(messages) > 0:
            if re.match("^\s*CC\s*(\[M\]\s*)?" + on_object, messages[0]):
                state = "CC"
                del messages[0]
                continue
            if re.match("^\s*CHECK\s*(\[M\]\s*)?" + on_file, messages[0]):
                state = "CHECK"
                del messages[0]
                continue
            if re.match("fixdep: [\S]* is empty", messages[0]):
                del messages[0]
                continue

            # Skip lines before "    CC"
            if state == None:
                pass
            elif state == "CC":
                CC.append(messages[0])
            elif state == "CHECK":
                CHECK.append(messages[0])
            else:
                raise RuntimeError("Should never been reached")

            # Remove line
            del messages[0]

        if statuscode != 0:
            logging.error("Running checker %s on file %s failed", extra_args, on_file)

        logging.debug("contents of CC:")
        logging.debug(CC)
        logging.debug("contents of CHECK:")
        logging.debug(CHECK)
        return (CC, CHECK)


    def call_gcc(self, on_file):
        """Call Gcc on the given file"""
        if "CC" in self.result_cache:
            return self.result_cache["CC"]

        extra_args = "KCFLAGS='%s'" % self.framework.options['args']['gcc']
        if self.framework.options.has_key('cross_prefix') and \
                len(self.framework.options['cross_prefix']) > 0:
            extra_args += " CROSS_COMPILE=%s" % self.framework.options['cross_prefix']
        (CC, _) = self.__call_make(on_file, extra_args)

        messages = GccMessage.preprocess_messages(CC)
        messages = [GccMessage(self, x) for x in messages]
        self.result_cache["CC"] = messages
        return messages


    def call_sparse(self, on_file):
        """Call Sparse on the given file"""
        if "SPARSE" in self.result_cache:
            return self.result_cache["SPARSE"]

        sparse = "ulimit -t 30; sparse"
        if 'sparse' in self.framework.options['args']:
            sparse += " " + self.framework.options['args']['sparse']

        (CC, CHECK) = self.__call_make(on_file, "C=2 CC='fakecc' CHECK='%s'" % sparse.replace("'", "\\'"))

        # GCC messages
        messages = GccMessage.preprocess_messages(CC)
        messages = map(lambda x: GccMessage(self, x), messages)
        self.result_cache["CC"] = messages

        # Sparse messages
        messages = SparseMessage.preprocess_messages(CHECK)
        messages = map(lambda x: SparseMessage(self, x), messages)
        self.result_cache["SPARSE"] = messages

        return messages


class LinuxPartialConfiguration(LinuxConfiguration):
    """
    This class creates a configuration object for a partial Linux
    Configuration. This works on arbitrary partial configurations, like
    "trolled" ones.

    NB: the self.cppflags and self.source is set to "/dev/null"
    """

    def __init__(self, framework, filename, arch, subarch,
                 expansion_strategy='alldefconfig'):
        assert arch
        logging.debug("arch: %s, subarch: %s", arch, subarch)
        LinuxConfiguration.__init__(self, framework,
                                    arch=arch, subarch=subarch,
                                    basename=filename, nth="",
                                    expansion_strategy=expansion_strategy)

        self.cppflags = '/dev/null'
        self.source   = '/dev/null'
        self.kconfig  = filename

    def write_config_h(self, dummy):
        pass

    def filename(self):
        return self.basename

    def save_expanded(self, config):
        # do not copy expanded configs around
        return '.config'


class LinuxStdConfiguration(LinuxConfiguration):
    """
    This class creates a configuration object for a standard Linux
    configuration, such as 'allyesconfig' or 'allnoconfig'.

    Instantiating this class will not change the current working tree,
    immediately.
    """

    def __init__(self, framework, basename, arch, subarch):
        assert framework.options.has_key('stdconfig')
        assert arch
        assert subarch
        logging.debug("arch: %s, subarch: %s", arch, subarch)
        configuration = ".%s" % framework.options['stdconfig']
        LinuxConfiguration.__init__(self, framework,
                                    arch=arch, subarch=subarch,
                                    basename=basename, nth=configuration)

        self.cppflags = '/dev/null'
        self.source   = '/dev/null'
        self.kconfig  = '/dev/null'

    def expand(self, verify=False):

        arch    = self.framework.options['arch']
        subarch = self.framework.options['subarch']

        call_linux_makefile(self.framework.options['stdconfig'],
                            arch=arch, subarch=subarch,
                            failok=False)

        if not self.framework.options.has_key('stdconfig_files'):
            self.framework.options['stdconfig_files'] \
                = set(files_for_current_configuration(arch, subarch))
        else:
            apply_configuration(arch=self.arch, subarch=self.subarch)
        self.kconfig = '.vampyr.' + self.framework.options['stdconfig']
        shutil.copy('.config', self.kconfig)

    def get_cppflags(self):
        return ""

    def filename(self):
        return self.basename + '.' + self.framework.options['stdconfig']



# Copyright (C) 2012 Valentin Rothberg <valentinrothberg@googlemail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

#!/usr/bin/env python

import unittest2 as t
import tempfile
from vamos.Config import Config

class TestConfig(t.TestCase):
    def setUp(self):
        self.configA = tempfile.NamedTemporaryFile()
        self.configB = tempfile.NamedTemporaryFile()

        self.configA.file.write("""CONFIG_A=y
CONFIG_B=y
CONFIG_C=n
CONFIG_M=m""")
        self.configA.file.flush()

        self.configB.file.write("""CONFIG_A=y
CONFIG_B=n
CONFIG_C=y
CONFIG_M=n""")
        self.configB.file.flush()

    def test_readConfigFile(self):
        confA = Config(self.configA.name)
        self.assertTrue(confA.contains("CONFIG_A"))

    def test_valueOf(self):
        confA = Config()
        confA.readConfigFile(self.configA.name)
        self.assertEqual(confA.valueOf("CONFIG_A"),"y")

    def test_addFeature(self):
        confA = Config()
        confA.addFeature("CONFIG_FOO","m")
        self.assertEqual(confA.valueOf("CONFIG_FOO"),"m")

    def test_conflict(self):
        confA = Config()
        confA.readConfigFile(self.configA.name)

        confB = Config()
        confB.readConfigFile(self.configB.name)

        self.assertTrue(confA.conflict(confB))

    def test_getConflicts(self):
        confA = Config()
        confA.readConfigFile(self.configA.name)

        confB = Config()
        confB.readConfigFile(self.configB.name)

        conflicts = confA.getConflicts(confB)
        self.assertIn("CONFIG_B",conflicts)
        self.assertIn("CONFIG_C",conflicts)
        self.assertIn("CONFIG_M",conflicts)

if __name__ == '__main__':
    t.main()



#
#   rsf2model - extracts presence implications from kconfig dumps
#
# Copyright (C) 2011 Christian Dietrich <christian.dietrich@informatik.uni-erlangen.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import functools
import sys

class UnicodeMixin(object):
    """3.0 compatible version of __str__"""
    if sys.version_info > (3, 0):
        __str__ = lambda x: x.__unicode__()
    else:
        __str__ = lambda x: unicode(x).encode('utf-8')

class Repr(UnicodeMixin):
    """Makes the representation of a object its str()"""
    __repr__ = lambda x: x.__unicode__()


class memoized(object):
    """Decorator that caches a function's return value each time it is called.
    If called later with the same arguments, the cached value is returned, and
    not re-evaluated.
    """
    def __init__(self, func):
        self.func = func
        self.cache = {}

    def __call__(self, *args):
        try:
            return self.cache[args]
        except KeyError:
            value = self.func(*args)
            self.cache[args] = value
            return value
        except TypeError:
            # uncachable -- for instance, passing a list as an argument.
            # Better to not cache than to blow up entirely.
            return self.func(*args)

    def __repr__(self):
        """Return the function's docstring."""
        return self.func.__doc__

    def __get__(self, obj, objtype):
        """Support instance methods."""
        return functools.partial(self.__call__, obj)


free_count = 0
def new_free_item():
    global free_count
    i = free_count
    free_count += 1
    return "__FREE__%d" % i



# Copyright (C) 2011 Christian Dietrich <christian.dietrich@informatik.uni-erlangen.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import unittest as t

from vamos.selection import Selection


class testSelection(t.TestCase):
    def setUp(self):
        self.empty = Selection()
        self.one_item = Selection([set(["CONFIG_PM"])])
        self.numa_item = Selection([set(["CONFIG_NUMA"])])
        self.two_item = Selection([set(["CONFIG_PM"]), set(["CONFIG_ACPI"])])
        self.two_item_module = Selection([set(["CONFIG_PM"]), set(["CONFIG_ACPI_MODULE"])])

        self.alternative = Selection([set(["CONFIG_PM"]), set(["CONFIG_ACPI", "CONFIG_SOUND"])])

        self.not_better_than_one_item = Selection([set(["CONFIG_X86"]), set(["CONFIG_SOUND"])])


    def test_equal(self):
        new = Selection(self.empty)
        self.assertEqual(str(self.empty), "")
        self.assertEqual(str(new), "")
        self.assertEqual(self.one_item, self.one_item)
        self.assertNotEqual(self.one_item, self.two_item)
        self.assertNotEqual(self.one_item, None)
        # can only compare to other selections
        self.assertNotEqual(Selection([set(["CONFIG_FOO"])]), "CONFIG_FOO")



    def test_to_str(self):
        self.assertEqual(str(self.one_item), "CONFIG_PM")
        self.assertEqual(str(self.two_item), "CONFIG_ACPI && CONFIG_PM")
        self.assertEqual(str(self.alternative), "(CONFIG_ACPI || CONFIG_SOUND) && CONFIG_PM")


    def test_better_than(self):
        self.assertTrue(self.one_item.better_than(self.two_item))
        self.assertTrue(self.one_item.better_than(self.alternative))

        self.assertFalse(self.one_item.better_than(self.not_better_than_one_item))
        self.assertFalse(self.two_item.better_than(self.one_item))

    def test_merge_selections(self):
        # Two unmergable expressions
        self.assertEqual(self.two_item.merge(self.not_better_than_one_item), None)

        merged = self.two_item.merge(self.two_item_module)
        self.assertNotEqual(merged, None)
        self.assertEqual(str(merged), "(CONFIG_ACPI || CONFIG_ACPI_MODULE) && CONFIG_PM")
        failedmerge = self.one_item.merge(self.two_item)
        self.assertEqual(failedmerge, None)

    def test_features(self):
        self.assertEqual(self.empty.features(), {})
        self.assertEqual(self.one_item.features(), {'CONFIG_PM': 'y'})
        self.assertEqual(self.two_item.features(), {'CONFIG_PM': 'y', 'CONFIG_ACPI':'y'})
        self.assertEqual(self.two_item_module.features(), {'CONFIG_PM': 'y', 'CONFIG_ACPI':'m'})

        alternate = self.alternative.features()
        self.assertTrue( alternate == {'CONFIG_PM': 'y', 'CONFIG_SOUND':'y'} or \
                         alternate == {'CONFIG_PM': 'y', 'CONFIG_ACPI':'y'})

        self.assertTrue(self.two_item_module.feature_in_selection("CONFIG_ACPI"))
        self.assertTrue(self.two_item_module.feature_in_selection("CONFIG_ACPI_MODULE"))

    def test_alternatives(self):
        newitem = self.one_item
        newitem.add_alternative("CONFIG_NUMA")
        self.assertEqual(str(newitem), "(CONFIG_NUMA || CONFIG_PM)")
        newitem.push_down()
        newitem.add_alternative("CONFIG_ACPI")
        self.assertEqual(str(newitem), "(CONFIG_NUMA || CONFIG_PM) && CONFIG_ACPI")

    def test_selection_invariance(self):
        a = Selection([set(["CONFIG_BARFOO"])])
        self.assertEqual(a.symbols, set(["CONFIG_BARFOO"]))

        self.assertEqual(self.two_item_module.symbols, set(["CONFIG_PM", "CONFIG_ACPI_MODULE"]))


if __name__ == '__main__':
    t.main()



#!/usr/bin/env python

import fnmatch
import os
import sys

import json

matches = []

result = {}
for parent, dirnames, filenames in os.walk('.'):
  for fn in fnmatch.filter(filenames, '*globally*'):
      filename = os.path.join(parent, fn)

      fd = open(filename)
      header = fd.readline().split(":")
      linecount = [header[2], header[5]]
      fd.close()

      path = filename.split("/")
      root = result
      for item in range(1, len(path) - 1):
          if not path[item] in root:
              root[path[item]] = {}
          root = root[path[item]]

      root[path[-1]] = linecount
      matches.append(filename)

print json.dumps(result)

sys.stderr.write("%d deads\n" % len(matches))



#!/usr/bin/env python

import glob, os, subprocess, sys

format_dead = "#line%(linum)s { background-color: #ff6655; cursor: pointer;}\n"
format_undead = "#line%(linum)s { background-color: #66dd11; cursor: pointer;}\n"

icon_file_css = """a.icon {
            padding-left: 1.5em;
            text-decoration: none;
     }
a.icon:hover {
     text-decoration: underline;
}
a.file {
            background : url("data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAABAAAAAQCAIAAACQkWg2AAAABnRSTlMAAAAAAABupgeRAAABHUlEQVR42o2RMW7DIBiF3498iHRJD5JKHurL+CRVBp+i2T16tTynF2gO0KSb5ZrBBl4HHDBuK/WXACH4eO9/CAAAbdvijzLGNE1TVZXfZuHg6XCAQESAZXbOKaXO57eiKG6ft9PrKQIkCQqFoIiQFBGlFIB5nvM8t9aOX2Nd18oDzjnPgCDpn/BH4zh2XZdlWVmWiUK4IgCBoFMUz9eP6zRN75cLgEQhcmTQIbl72O0f9865qLAAsURAAgKBJKEtgLXWvyjLuFsThCSstb8rBCaAQhDYWgIZ7myM+TUBjDHrHlZcbMYYk34cN0YSLcgS+wL0fe9TXDMbY33fR2AYBvyQ8L0Gk8MwREBrTfKe4TpTzwhArXWi8HI84h/1DfwI5mhxJamFAAAAAElFTkSuQmCC ") left top no-repeat;
          }
a.dir {
            background : url("data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAABAAAAAQCAYAAAAf8/9hAAAAGXRFWHRTb2Z0d2FyZQBBZG9iZSBJbWFnZVJlYWR5ccllPAAAAd5JREFUeNqMU79rFUEQ/vbuodFEEkzAImBpkUabFP4ldpaJhZXYm/RiZWsv/hkWFglBUyTIgyAIIfgIRjHv3r39MePM7N3LcbxAFvZ2b2bn22/mm3XMjF+HL3YW7q28YSIw8mBKoBihhhgCsoORot9d3/ywg3YowMXwNde/PzGnk2vn6PitrT+/PGeNaecg4+qNY3D43vy16A5wDDd4Aqg/ngmrjl/GoN0U5V1QquHQG3q+TPDVhVwyBffcmQGJmSVfyZk7R3SngI4JKfwDJ2+05zIg8gbiereTZRHhJ5KCMOwDFLjhoBTn2g0ghagfKeIYJDPFyibJVBtTREwq60SpYvh5++PpwatHsxSm9QRLSQpEVSd7/TYJUb49TX7gztpjjEffnoVw66+Ytovs14Yp7HaKmUXeX9rKUoMoLNW3srqI5fWn8JejrVkK0QcrkFLOgS39yoKUQe292WJ1guUHG8K2o8K00oO1BTvXoW4yasclUTgZYJY9aFNfAThX5CZRmczAV52oAPoupHhWRIUUAOoyUIlYVaAa/VbLbyiZUiyFbjQFNwiZQSGl4IDy9sO5Wrty0QLKhdZPxmgGcDo8ejn+c/6eiK9poz15Kw7Dr/vN/z6W7q++091/AQYA5mZ8GYJ9K0AAAAAASUVORK5CYII= ") left top no-repeat;           }

a.up {   background : url("data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAABAAAAAQCAYAAAAf8/9hAAAAGXRFWHRTb2Z0d2FyZQBBZG9iZSBJbWFnZVJlYWR5ccllPAAAAmlJREFUeNpsU0toU0EUPfPysx/tTxuDH9SCWhUDooIbd7oRUUTMouqi2iIoCO6lceHWhegy4EJFinWjrlQUpVm0IIoFpVDEIthm0dpikpf3ZuZ6Z94nrXhhMjM3c8895977BBHB2PznK8WPtDgyWH5q77cPH8PpdXuhpQT4ifR9u5sfJb1bmw6VivahATDrxcRZ2njfoaMv+2j7mLDn93MPiNRMvGbL18L9IpF8h9/TN+EYkMffSiOXJ5+hkD+PdqcLpICWHOHc2CC+LEyA/K+cKQMnlQHJX8wqYG3MAJy88Wa4OLDvEqAEOpJd0LxHIMdHBziowSwVlF8D6QaicK01krw/JynwcKoEwZczewroTvZirlKJs5CqQ5CG8pb57FnJUA0LYCXMX5fibd+p8LWDDemcPZbzQyjvH+Ki1TlIciElA7ghwLKV4kRZstt2sANWRjYTAGzuP2hXZFpJ/GsxgGJ0ox1aoFWsDXyyxqCs26+ydmagFN/rRjymJ1898bzGzmQE0HCZpmk5A0RFIv8Pn0WYPsiu6t/Rsj6PauVTwffTSzGAGZhUG2F06hEc9ibS7OPMNp6ErYFlKavo7MkhmTqCxZ/jwzGA9Hx82H2BZSw1NTN9Gx8ycHkajU/7M+jInsDC7DiaEmo1bNl1AMr9ASFgqVu9MCTIzoGUimXVAnnaN0PdBBDCCYbEtMk6wkpQwIG0sn0PQIUF4GsTwLSIFKNqF6DVrQq+IWVrQDxAYQC/1SsYOI4pOxKZrfifiUSbDUisif7XlpGIPufXd/uvdvZm760M0no1FZcnrzUdjw7au3vu/BVgAFLXeuTxhTXVAAAAAElFTkSuQmCC ") left top no-repeat;      }
"""

def dead_text(count):
    if count == 0:
        return ""
    return str(count)

def dead_file(filename, version):
    html = open(filename + ".html", "w+")
    deads = glob.glob(filename+"*globally*dead") + glob.glob(filename + "*globally*undead")

    p = subprocess.Popen("code2html -n " + filename, shell = True,
                         stdin=subprocess.PIPE, stdout=html, stderr=subprocess.PIPE)

    p.stdin.write('<script type="text/javascript" src="/Research/VAMOS/linux-trees/%s/jquery.js"></script>\n' % version);
    p.stdin.write('<script type="text/javascript" src="/Research/VAMOS/linux-trees/%s/explorer.js"></script>\n' % version);

    p.stdin.close()
    p.wait()
    return len(deads)

def dead_directory(dir=".", version = ""):
    fd = open(os.path.join(dir, "index.html"), "w+")
    deads = 0


    fd.write("""<html>
<head>
<style type="text/css">
""" + icon_file_css + """
</style>
</head>
<body>""")
    fd.write("<h1>Index for /%s</h1><hr>" % dir.replace("./", ""))
    fd.write("<table>")
    fd.write("<tr><td></td><td><a href='../index.html' class='icon up'>[parent directory]</a></td></tr>\n")

    paths = [path for path in os.listdir(dir) if os.path.isdir(os.path.join(dir, path))]
    paths_ = [os.path.join(dir, path) for path in paths]
    counts = map(lambda x:dead_directory(x, version), paths_)
    deads += sum(counts)
    for (path, d) in zip(paths, counts):
            fd.write("<tr><td>%s</td><td><a href='%s/index.html' class='icon dir'>%s</a></td></tr>\n" %(dead_text(d), path, path))


    files = glob.glob(dir + "/*.h") + glob.glob(dir + "/*.c")
    counts = map(lambda x: dead_file(x, version), files)
    deads += sum(counts)
    for (filename, d) in zip(files, counts):
        f = os.path.basename(filename)
        fd.write("<tr><td>%s</td><td><a href='%s.html' class='icon file'>%s</a></td></tr>\n" %(dead_text(d), f, f))

    fd.write("</ul>")

    fd.close()
    return deads

if __name__ == '__main__':
    print dead_directory(sys.argv[1], sys.argv[2])



#
#   golem - analyzes feature dependencies in Linux makefiles
#
# Copyright (C) 2011 Christian Dietrich <christian.dietrich@informatik.uni-erlangen.de>
# Copyright (C) 2012 Reinhard Tartler <tartler@informatik.uni-erlangen.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from vamos.golem.kbuild import files_for_selected_features
import logging

class FileSet:
    """
    A FileSet is a list of files and directories, compiled by a given selection.
    """

    def __init__(self, selection, arch, subarch=None):
        self.arch = arch
        self.subarch = subarch
        self.selection = selection
        try:
            (self.files, self.dirs) = files_for_selected_features(selection.features(), arch, subarch)
        except RuntimeError as e:
            logging.error("Failed to determine fileset, continuing anyways: %s", e)
            self.files = set()
            self.dirs = set()
            return
        self.files = self.files
        self.dirs = self.dirs

    def compare_to_base(self, other):
        """
        Compare the filesets to a other selection.

        @return ((f_added, f_deleted), (d_added, d_deleted))
        """

        f_added = self.files - other.files
        f_deleted = other.files - self.files
        d_added = self.dirs - other.dirs
        d_deleted = other.dirs - self.dirs

        return ((f_added, f_deleted), (d_added, d_deleted))

class FileSetCache(dict):
    def __init__(self, arch, subarch):
        dict.__init__(self)
        self.arch = arch
        self.subarch = subarch

    def get_fileset(self, selection):
        if not selection in self:
            self[selection] = [FileSet(selection, arch=self.arch, subarch=self.subarch), 1]
        else:
            # hit
            # logging.info("HIT " + str(sys.getsizeof(self)) + " " + str(len(self)))
            self[selection][1] += 1
        return self[selection][0]





#
#   rsf2model - extracts presence implications from kconfig dumps
#
# Copyright (C) 2011 Christian Dietrich <christian.dietrich@informatik.uni-erlangen.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from vamos.rsf2model import tools
from vamos.rsf2model.RsfReader import Choice
from vamos.rsf2model import BoolRewriter
from vamos.rsf2model.helper import BoolParserException

class TranslatedModel(tools.UnicodeMixin):
    def __init__(self, rsf):
        tools.UnicodeMixin.__init__(self)

        self.symbols = []
        self.deps = {}
        self.defaultSelects = {}

        self.always_on = []

        self.rsf = rsf
        for t in self.rsf.options().items():
            option = t[1]
            self.translate_option(option)

        for t in self.rsf.options().items():
            option = t[1]
            if type(option) == Choice:
                self.translate_choice(option)

        for (item, default_set) in self.rsf.collect("Default", 0, True).items():
            for default in default_set:
                option = self.rsf.options().get(item, None)
                if option:
                    try:
                        self.translate_default(option, default)
                    except BoolParserException:
                        # Parsing expression failed, just ignore it
                        pass
        for (item, select_set) in self.rsf.collect("ItemSelects", 0, True).items():
            for select in select_set:
                option = self.rsf.options().get(item, None)
                if option:
                    try:
                        self.translate_select(option, select)
                    except BoolParserException:
                        # Parsing of a substring failed, just ignore it
                        pass

    def translate_option(self, option):
        # Generate symbols
        symbol = option.symbol()
        self.symbols.append(symbol)
        self.deps[symbol] = []
        self.defaultSelects[symbol] = []

        if option.omnipresent():
            self.always_on.append(option.symbol())

        if option.tristate():
            symbol_module = option.symbol_module()
            self.symbols.append(symbol_module)
            self.deps[symbol_module] = []

            # mutual exclusion for tristate symbols
            self.deps[symbol].append("!%s" % symbol_module)
            self.deps[symbol_module].append("!%s" % symbol)
            self.deps[symbol_module].append("CONFIG_MODULES")

        # Add dependency
        if option.tristate():
            # For tristate symbols (without _MODULE) the dependency must evaluate to y
            dep = option.dependency(eval_to_module = False)
            if dep:
                self.deps[symbol].insert(0, dep)

            # for _MODULE also "m" is ok
            dep = option.dependency(eval_to_module = True)
            if dep:
                self.deps[option.symbol_module()].insert(0, dep)
        else:
            dep = option.dependency(eval_to_module = True)
            if dep:
                self.deps[symbol].insert(0, dep)

    def translate_choice(self, choice):
        for (symbol, dep) in choice.insert_forward_references().items():
            self.deps[symbol].extend(dep)

    def translate_default(self, option, dependency):
        # dependency: [state, if <expr>]
        if len(dependency) != 2:
            return

        if type(option) == Choice or option.tristate() or option.prompts() != 0:
            return

        [state, cond] = dependency
        if state == "y" and cond == "y" \
               and option.prompts() == 0 \
               and not option.tristate():
            self.always_on.append(option.symbol())
            # we add a free item to the defaultSelect list. this is
            # important in the following scenario:
            # 1. Item ist default on
            # 2. item is selected by other item
            # 3. without the free item it is implied, that the second
            #    item is on, byt with default to y this isn't right
            #
            # Difference:
            # A DEPS && (SELECT1)
            # A DEPS && (__FREE__ || SELECT1)
            self.defaultSelects[option.symbol()].append(tools.new_free_item())
        elif state == "y" or cond == "y":
            expr = state
            if state == "y":
                expr = cond
            expr =  str(BoolRewriter.BoolRewriter(self.rsf, expr, eval_to_module = True).rewrite())
            self.defaultSelects[option.symbol()].append(expr)
        # Default FOO "BAR" "BAZ"
        elif len(state) > 1 and len(cond) > 1:
            expr =  str(BoolRewriter.BoolRewriter(self.rsf, "(%s) && (%s)" % (state, cond),
                                                  eval_to_module = True).rewrite())
            self.defaultSelects[option.symbol()].append(expr)


    def translate_select(self, option, select):
        """
        @param option the option that declares the 'select' verb
        @param select: the option that is unconditionally slected (if option is selected

        select is a tuple of (seleted, condition)
        """
        if type(option) == Choice:
            return

        if len(select) != 2:
            return

        (selected, expr) = select

        selected = self.rsf.options().get(selected, None)
        if not selected:
            return

        # We only look on boolean selected options
        if type(selected) == Choice or selected.tristate():
            return

        if expr == "y":
            # select foo if y
            if selected.prompts() == 0:
                self.defaultSelects[selected.symbol()].append(option.symbol())
            self.deps[option.symbol()].append(selected.symbol())

            if option.tristate():
                if selected.prompts() == 0:
                    self.defaultSelects[selected.symbol()].append(option.symbol_module())
                self.deps[option.symbol_module()].append(selected.symbol())
        else:
            # select foo if expr
            expr = str(BoolRewriter.BoolRewriter(self.rsf, expr, eval_to_module = True).rewrite())
            if expr == "":
                # We are in a Choice, and the choice is the only dependency....
                imply = selected.symbol()
            else:
                imply = "((%s) -> %s)" %(expr, selected.symbol())

            if selected.prompts() == 0:
                self.defaultSelects[selected.symbol()].append(option.symbol())

            self.deps[option.symbol()].append(imply)
            if option.tristate():
                if selected.prompts() == 0:
                    self.defaultSelects[selected.symbol()].append(option.symbol_module())
                self.deps[option.symbol_module()].append(imply)


    def __unicode__(self):
        result = u""
        result += u"I: Items-Count: %d\n" % len(self.symbols)
        result += u"I: Format: <variable> [presence condition]\n"
        result += "UNDERTAKER_SET SCHEMA_VERSION 1.1\n"

        if len(self.always_on) > 0:
            result += "UNDERTAKER_SET ALWAYS_ON " + (" ".join(map(lambda x: '"' + x + '"', self.always_on))) + "\n"

        for symbol in self.symbols:
            expression = ""
            deps = self.deps.get(symbol, [])
            if symbol in self.defaultSelects and len(self.defaultSelects[symbol]) > 0:
                dS = self.defaultSelects[symbol]
                deps.append("(" + (" || ".join(dS)) + ")")

            expression = " && ".join(deps)
            if expression == "":
                result += "%s\n" % symbol
            else:
                result += "%s \"%s\"\n" % (symbol, expression)
        return result



#!/usr/bin/env python
#
# Copyright (C) 2011 Christian Dietrich <christian.dietrich@informatik.uni-erlangen.de>
# Copyright (C) 2011 Valentin Rothberg <valentinrothberg@googlemail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import unittest2 as t
from vamos.model import Model
import tempfile

class TestModel(t.TestCase):
    def setUp(self):
        self.model_file = tempfile.NamedTemporaryFile()
        self.model_file.file.write("""I: Items-Count: 12399
I: Format: <variable> [presence condition]
UNDERTAKER_SET SCHEMA_VERSION 1.1
UNDERTAKER_SET ALWAYS_ON "CONFIG_B1 "CONFIG_B2"
CONFIG_T1
CONFIG_B3 "CONFIG_T2 -> CONFIG_FOO && (CONFIG_BAR)"
CONFIG_B4 CONFIG_T1
CONFIG_T2 "!CONFIG_T2_MODULE && CONFIG_FOO"
CONFIG_T2_MODULE "!CONFIG_T2 && CONFIG_BAR && CONFIG_MODULES"
CONFIG_T3 "CONFIG_BAR"
""")
        self.model_file.file.flush()

        self.rsf_file = tempfile.NamedTemporaryFile()
        self.rsf_file.file.write("""Item B1 boolean
Default "y"
Item B2 boolean
Default "y"
Item T1 tristate
Default T1 "y"
Item B3 boolean
HasPrompts B3 0
Default B3 "FOO && BLA" "y"
Item B4 boolean
Default B4 "A" "y"
Item T2 tristate
Depends T2 "BAR"
Item T3 tristate
Depends T3 "BAR"
""")
        self.rsf_file.file.flush()

    def test_all_symbols_included(self):
        model = Model(self.model_file.name)
        for symbol in ("CONFIG_B1", "CONFIG_B2", "CONFIG_B3", "CONFIG_B4",
                       "CONFIG_T1", "CONFIG_T2_MODULE", "CONFIG_T2", "CONFIG_T3"):
            self.assertIn(symbol, model)

        for symbol in ("CONFIG_ALWAYS_ON", "ALWAYS_ON"):
            self.assertNotIn(symbol, model)

    def test_mentioned_items(self):
        model = Model(self.model_file.name)
        self.assertEqual(model.mentioned_items("CONFIG_T1"), [])
        self.assertEqual(model.mentioned_items("CONFIG_FOO"), [])
        self.assertEqual(model.mentioned_items("CONFIG_B4"), ["CONFIG_T1"])
        self.assertEqual(set(model.mentioned_items("CONFIG_B3")), set(["CONFIG_T2",
                "CONFIG_FOO", "CONFIG_BAR"]))
        self.assertEqual(set(model.mentioned_items("CONFIG_T2")), set(["CONFIG_T2_MODULE",
                "CONFIG_FOO"]))
        self.assertEqual(set(model.mentioned_items("CONFIG_T2_MODULE")), set(["CONFIG_T2",
                "CONFIG_MODULES","CONFIG_BAR"]))
        self.assertEqual(set(model.mentioned_items("CONFIG_T3")), set(["CONFIG_BAR"]))

    def test_symbol_always_on(self):
        model = Model(self.model_file.name)
        for symbol in ("CONFIG_ALWAYS_ON", "CONFIG_BARFOO",
                       "CONFIG_A", "CONFIG_B", "CONFIG_C", "CONFIG_MODULE"):
            self.assertNotIn(symbol, model.always_on_items)

        for symbol in ("CONFIG_B1", "CONFIG_B2"):
            self.assertIn(symbol, model.always_on_items)

    def test_slice_symbols(self):
        model = Model(self.model_file.name)
        self.assertEqual(set(model.slice_symbols(["CONFIG_B1"])),set(["CONFIG_B1"]))
        self.assertEqual(set(model.slice_symbols(["CONFIG_B2"])),set(["CONFIG_B2"]))
        self.assertEqual(set(model.slice_symbols(["CONFIG_B3"])),set(["CONFIG_B3",
                "CONFIG_T2", "CONFIG_FOO", "CONFIG_BAR", "CONFIG_T2_MODULE",
                "CONFIG_MODULES"]))
        self.assertEqual(set(model.slice_symbols(["CONFIG_B4"])),set(["CONFIG_B4",
                "CONFIG_T1"]))
        self.assertEqual(set(model.slice_symbols(["CONFIG_T1"])),set(["CONFIG_T1"]))
        self.assertEqual(set(model.slice_symbols(["CONFIG_T2"])),set(["CONFIG_T2",
                "CONFIG_T2_MODULE", "CONFIG_FOO", "CONFIG_BAR", "CONFIG_MODULES"]))
        self.assertEqual(set(model.slice_symbols(["CONFIG_T2_MODULE"])),
                set(["CONFIG_T2_MODULE", "CONFIG_T2", "CONFIG_BAR", "CONFIG_MODULES",
                "CONFIG_FOO"]))

    def test_leaf_features(self):
        model = Model(self.model_file.name)
        for x in ("CONFIG_B1", "CONFIG_B2", "CONFIG_B3", "CONFIG_B4", "CONFIG_T3"):
            self.assertIn(x, model.leaf_features())

    def test_none_leaf_features(self):
        model = Model(self.model_file.name)
        self.assertNotIn(set(["CONFIG_FOO", "CONFIG_BAR", "CONFIG_T1", "CONFIG_T2"]),
                set(model.leaf_features()))

    def test_types(self):
        model = Model(self.model_file.name, self.rsf_file.name)
        for x in ("CONFIG_B1", "CONFIG_B2", "CONFIG_B3", "CONFIG_B4"):
            self.assertEqual(model.get_type(x), "boolean")
        for x in ("CONFIG_T1", "CONFIG_T2"):
            self.assertEqual(model.get_type(x), "tristate")



if __name__ == '__main__':
    t.main()



#!/usr/bin/env python
#
#   rsf2model - extracts presence implications from kconfig dumps
#
# Copyright (C) 2011 Christian Dietrich <christian.dietrich@informatik.uni-erlangen.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import unittest as t
import StringIO
from vamos.rsf2model import RsfReader


class TestRsfReader(t.TestCase):
    def setUp(self):
        rsf = """Item A 'boolean'
Item B tristate
ItemFoo B some_value
Item C boolean
Item H hex
Item S string
CRAP CARASDD"""
        self.rsf = RsfReader.RsfReader(StringIO.StringIO(rsf))

    def test_correct_options(self):
        opts = set()
        for name in self.rsf.options():
            opt = self.rsf.options()[name]
            if opt.name == "A":
                opts.add("A")
                self.assertEqual(opt.tristate(), False, "Item state wrong parsed")
            if opt.name == "B":
                opts.add("B")
                self.assertEqual(opt.tristate(), True, "Item state wrong parsed")
            if opt.name == "C":
                opts.add("C")
                self.assertEqual(opt.tristate(), False, "Item state wrong parsed")
            if opt.name == "H":
                opts.add("H")
                self.assertEqual(opt.hex(), True, "Item state wrong")
            if opt.name == "S":
                opts.add("S")
                self.assertEqual(opt.string(), True, "Item state wrong")

    def test_symbol_generation(self):
        self.assertEqual(self.rsf.options()["A"].symbol(), "CONFIG_A")
        self.assertRaises(RsfReader.OptionNotTristate, self.rsf.options()["A"].symbol_module)
        self.assertEqual(self.rsf.options()["B"].symbol_module(), "CONFIG_B_MODULE")
        self.assertEqual(self.rsf.options()["S"].symbol(), "CONFIG_S")
        self.assertEqual(self.rsf.options()["H"].symbol(), "CONFIG_H")


if __name__ == '__main__':
    t.main()



#
#   rsf2model - extracts presence implications from kconfig dumps
#
# Copyright (C) 2012 Manuel Zerpies <manuel.f.zerpies@ww.stud.uni-erlangen.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import ast
import re
from vamos.rsf2model.helper import BoolParserException

class BoolParser (ast.NodeTransformer):
    AND = "and"
    OR  = "or"
    NOT = "not"
    EQUAL = "=="
    NEQUAL = "!="

    def __init__(self, bool_expr):
        ast.NodeTransformer.__init__(self)

        try:
            expr = bool_expr.replace("&&", " and ").replace("||", "or")
            expr = re.sub("!(?=[^=])", " not ", expr)
            expr = re.sub("(?<=[^!])=", " == ", expr)
            expr = re.sub("((?<=[( &|])|^)(?=[0-9])", "VAMOS_MAGIC_", expr)
            expr = re.sub("^ *", "", expr)
            self.tree = ast.parse(expr)
        except:
            raise BoolParserException("Parsing failed: '" + expr + "'")

        if self.tree.__class__ != ast.Module \
                or len(self.tree.body) > 1 \
                or self.tree.body[0].__class__ != ast.Expr:
            raise BoolParserException("No valid boolean expression")
        self.tree = self.tree.body[0]

    @staticmethod
    def new_node(node_type, childs):
        return [node_type] + childs

    def visit_Expr(self, node):
        return self.visit(node.value)

    def visit_BoolOp(self, node):
        if type(node.op) == ast.And:
            return self.new_node(self.AND, map(self.visit, node.values))
        if type(node.op) == ast.Or:
            return self.new_node(self.OR,  map(self.visit, node.values))
        raise BoolParserException("Unkown boolean expression")

    def visit_UnaryOp(self, node):
        if type(node.op) == ast.Not:
            return self.new_node(self.NOT, [self.visit(node.operand)])
        raise BoolParserException("Unkown unary operation")

    def visit_Compare(self, node):
        if len(node.ops) == 1 and len(node.comparators) == 1:
            if type(node.ops[0]) in [ast.Eq, ast.NotEq]:
                left = self.visit(node.left)
                right = self.visit(node.comparators[0])
                if type(left) != str or type(right) != str:
                    raise BoolParserException("Too complex comparison")
                if type(node.ops[0]) == ast.Eq:
                    return self.new_node(self.EQUAL, [left, right])
                else:
                    return self.new_node(self.NEQUAL, [left, right])
        raise BoolParserException("Unkown compare operation")

    @staticmethod
    def visit_Name(node):
        if node.id.startswith("VAMOS_MAGIC_"):
            return node.id[len("VAMOS_MAGIC_"):]
        return node.id

    def visit(self, node):
        func = "visit_" + node.__class__.__name__
        if hasattr(self, func):
            return (getattr(self, func))(node)
        else:
            raise BoolParserException("Boolean Parsing failed, unkown expression: " + repr(node))
    def to_bool(self):
        return self.visit(self.tree)



#
#   BuildFrameworks - utility classes for working in source trees
#
# Copyright (C) 2011 Christian Dietrich <christian.dietrich@informatik.uni-erlangen.de>
# Copyright (C) 2011-2012 Reinhard Tartler <tartler@informatik.uni-erlangen.de>
# Copyright (C) 2012 Christoph Egger <siccegge@informatik.uni-erlangen.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from vamos.vampyr.Configuration \
    import BareConfiguration, LinuxConfiguration, LinuxStdConfiguration
from vamos.vampyr.utils import find_configurations, \
    get_conditional_blocks, get_block_to_linum_map
from vamos.golem.kbuild import apply_configuration, file_in_current_configuration, \
    guess_arch_from_filename, guess_subarch_from_arch, \
    call_linux_makefile, find_autoconf, \
    get_linux_version, NotALinuxTree
from vamos.tools import execute

import os.path
import logging

class BuildFramework:
    """ Base class for all Build System Frameworks"""

    def __init__(self, options=None):
        """
        options: Dictonary that control the used helper tools. The
        following keys are supported:

        args: Dictionary with extra arguments for various tools (see note below)

        exclude_others: boolean, if set, #include statements will be ignored.

        loglevel: sets the loglevel, see the logging package.

        Note: in the args dictionary, the 'undertaker' key can be used
        to pass extra arguments to undertaker when calculating the
        configurations. For instance, pass "-C simple" to select the
        non-greedy algorithm for caluculating configurations.

        For the tools 'sparse', 'gcc' and 'clang', this field allows you
        to enable additional warnings (e.g., -Wextra).
        """

        if options:
            assert(isinstance(options, dict))
            if options.has_key('args'):
                assert(isinstance(options['args'], dict))
            self.options = options
        else:
            self.options = None

    def calculate_configurations(self, filename):
        raise NotImplementedError

    def analyze_configuration_coverage(self, filename):
        """
        Analyzes the given file 'filename' for its configuration coverage.

        This method also includes statistics about blocks covered by 'allyeconfig'

        returns a dict with the following members:
         - lines_total: the total number of LOC in the file
         - lines_covered: the number of covered LOC in all configurations
         - lines_conditional_total: the number of LOC in #ifdef blocks
         - lines_conditional_covered: the number of LOC in selected #ifdef blocks
         - linum_map: the map as returned by 'get_block_to_linum_map(filename)'
         - blocks: a dict of configuration to the set of cond. blocks it contains
         - blocks_total: contains a set of all blocks in the file
         - blocks_covered: contains the set of all covered blocks over all configurations
         - blocks_uncovered: contains the set of blocks that are not in any configuration

        ..note: the dict blocks contains the configuration name as
                key. The name has the form 'configN', with N being a
                running number. The length if this dict is equal to the
                number of found confiurations of this file.

        """

        return_dict = {
            'blocks': {},
            'lines': {}}

        total_blocks = set(get_conditional_blocks(filename, None,
                                                  configuration_blocks=False))
        total_blocks.add("B00")
        return_dict['blocks_total'] = total_blocks

        conf_blocks = set(get_conditional_blocks(filename, None,
                                                 configuration_blocks=True))
        conf_blocks.add("B00")
        return_dict['configuration_blocks'] = conf_blocks

        covered_blocks = set()

        if self.options.has_key('configfile'):
            config_h = self.options['configfile']
            logging.info("Analyzing Configuration %s", config_h)
            covered_blocks = return_dict['blocks'][config_h] = \
                set(get_conditional_blocks(filename, config_h)) | set(["B00"])
        else:
            configs = self.calculate_configurations(filename)
            logging.info("%s: found %d configurations", filename, len(configs))

            return_dict['all_configs'] = find_configurations(filename)
            return_dict['all_config_count'] = len(return_dict['all_configs'])
            return_dict['expandable_configs'] = len(configs)

            for config in configs:
                try:
                    config.switch_to()
                    autoconf_h=config.get_config_h()
                    old_len = len(covered_blocks)
                    return_dict['blocks'][config.kconfig] = \
                        set(get_conditional_blocks(filename, autoconf_h,
                                                   configuration_blocks=False)) \
                                                   | set(["B00"])
                    covered_blocks |= return_dict['blocks'][config.kconfig]

                    print "Config %s added %d additional blocks" % \
                        (config.kconfig, len(covered_blocks) - old_len)

                except RuntimeError as e:
                    logging.error("Failed to process config '%s': %s", filename, e)

        return_dict['blocks_covered'] = covered_blocks
        return_dict['blocks_uncovered'] = total_blocks - covered_blocks

        linum_map = get_block_to_linum_map(filename)
        return_dict['linum map'] = linum_map

        total_lines = sum(linum_map.values())
        covered_lines = 0
        for block in covered_blocks:
            covered_lines += linum_map[block]

        return_dict['lines_total'] = total_lines
        return_dict['lines_covered'] = covered_lines
        return_dict['lines_conditional total'] = total_lines - linum_map["B00"]
        if "B00" in covered_blocks:
            return_dict['lines_conditional covered'] = covered_lines - linum_map["B00"]
        else:
            return_dict['lines_conditional covered'] = 0

        return return_dict


class BareBuildFramework(BuildFramework):
    """ For use without Makefiles, works directly on source files """

    def __init__(self, options=None):
        BuildFramework.__init__(self, options)

    def calculate_configurations(self, filename):
        undertaker = "undertaker -q -j coverage -C min -O combined"
        if 'undertaker' in self.options['args']:
            undertaker += " " + self.options['args']['undertaker']
        undertaker += " '" + filename + "'"
        execute(undertaker, failok=False)

        configs = list()
        for cfgfile in find_configurations(filename):
            # here, we rely on the fact that find_configurations returns filenames such as
            # 'kernel/sched.c.config42' (i.e., filenames end on the number)
            assert '.config' in cfgfile
            basename, ext = os.path.splitext(cfgfile)
            nth = int(ext[len('.config'):])
            configs.append(BareConfiguration(self, basename, nth))
        return configs


class KbuildBuildFramework(BuildFramework):
    """ For use with Linux, considers Kconfig constraints """

    def __init__(self, options=None):
        if not options:
            options=dict()
        BuildFramework.__init__(self, options)
        if not options.has_key('expansion_strategy'):
            options['expansion_strategy'] = 'alldefconfig'
        if not options.has_key('coverage_strategy'):
            options['coverage_strategy'] = 'min'
        if not options.has_key('arch'):
            options['arch'] = None
        if not options.has_key('subarch'):
            options['subarch'] = None

    def guess_arch_from_filename(self, filename):
        """
        Tries to guess the architecture from a filename. Uses golem's
        guess_arch_from_filename() method.
        """
        (arch, subarch) = guess_arch_from_filename(filename)
        self.options['arch'] = arch
        self.options['subarch'] = subarch
        return (arch, subarch)

    def calculate_configurations(self, filename):
        """Calculate configurations for a given file with Kconfig output mode"""

        if self.options['arch'] is None:
            oldsubarch = self.options['subarch']
            self.guess_arch_from_filename(filename)
            if oldsubarch is not None:
                # special case: if only subarch is set, restore it
                self.options['subarch'] = oldsubarch

        cmd = "undertaker -q -j coverage -C %s -O combined" % self.options['coverage_strategy']
        if os.path.isdir("models"):
            cmd += " -m models/%s.model" % self.options['arch']
        else:
            logging.info("No models directory found, running without models")

        logging.info("Calculating configurations for '%s'", filename)
        if self.options and self.options.has_key('args'):
            if 'undertaker' in self.options['args']:
                cmd += " " + self.options['args']['undertaker']

        cmd += " '%s'" % filename.replace("'", "\\'")
        (output, statuscode) = execute(cmd, failok=True)
        if statuscode != 0 or any([l.startswith("E:") for l in output]):
            logging.error("Running undertaker failed: %s", cmd)
            print "--"
            for i in output:
                logging.error(i)
            return set()

        logging.info("Testing which configurations are actually being compiled")
        configs = list()
        for cfgfile in find_configurations(filename):
            # here, we rely on the fact that find_configurations returns filenames such as
            # 'kernel/sched.c.config42' (i.e., filenames end on the number)
            assert '.config' in cfgfile
            basename, ext = os.path.splitext(cfgfile)
            nth = int(ext[len('.config'):])
            config_obj = LinuxConfiguration(self, basename, nth,
                                            arch=self.options['arch'], subarch=self.options['subarch'],
                                            expansion_strategy=self.options['expansion_strategy'])
            config_obj.switch_to()
            if file_in_current_configuration(filename, config_obj.arch, config_obj.subarch) != "n":
                logging.info("Configuration '%s' is actually compiled", cfgfile)
                configs.append(config_obj)
            else:
                logging.info("Configuration '%s' is *not* compiled", cfgfile)

        if self.options.has_key('stdconfig'):
            if self.options.has_key('stdconfig_files') \
                    and not filename in self.options['stdconfig_files']:
                logging.info("Skipping %s because stdconfig '%s' does not build it",
                             filename, self.options['stdconfig'])
                return configs

            c  = LinuxStdConfiguration(self, basename=filename,
                                       arch=self.options['arch'],
                                       subarch=self.options['subarch'])
            c.switch_to()
            if file_in_current_configuration(filename, c.arch, c.subarch) != "n":
                logging.info("Also including configuration '%s'", self.options['stdconfig'])
                configs.append(c)
            else:
                logging.info("configuration '%s' does not build '%s'",
                             self.options['stdconfig'], filename)

        return configs


    def analyze_configuration_coverage(self, filename):
        """
        Analyzes the given file 'filename' for its configuration coverage.

        This method also includes statistics about blocks covered by 'allyeconfig'

        returns a dict with the following members:
         - arch: the analyzed architecture
         - subarch: the analyzed subarchitecture
         - lines_total: the total number of LOC in the file
         - lines_covered: the number of covered LOC in all configurations
         - lines_conditional_total: the number of LOC in #ifdef blocks
         - lines_conditional_covered: the number of LOC in selected #ifdef blocks
         - linum_map: the map as returned by 'get_block_to_linum_map(filename)'
         - blocks: a dict of configuration to the set of cond. blocks it contains
         - blocks_allyesconfig: contains the set of blocks 'allyesconfig' contains
         - blocks_total: contains a set of all blocks in the file
         - blocks_covered: contains the set of all covered blocks over all configurations
         - blocks_uncovered: contains the set of blocks that are not in any configuration

        ..note: the dict blocks contains the configuration name as
                key. The name has the form 'configN', with N being a
                running number. The length if this dict is equal to the
                number of found confiurations of this file.

        """

        if self.options['arch'] is None:
            oldsubarch = self.options['subarch']
            self.guess_arch_from_filename(filename)
            if oldsubarch is not None:
                # special case: if only subarch is set, restore it
                self.options['subarch'] = oldsubarch

        # sanity check: remove existing configuration to ensure consistent behavior
        if os.path.exists(".config"):
            os.unlink(".config")
        apply_configuration(arch=self.options['arch'], subarch=self.options['subarch'])
        assert os.path.exists('.config')
        autoconf_h=find_autoconf()

        return_dict = BuildFramework.analyze_configuration_coverage(self, filename)

        return_dict['arch'] = self.options['arch']
        return_dict['subarch'] = self.options['subarch']


        if self.options.has_key('configfile'):
            logging.info("Skipping analysis of allyesconfig, since we analyze %s",
                         self.options['configfile'])

        # generate the configuration for 'allyesconfig'
        call_linux_makefile("allyesconfig", failok=False,
                            arch=self.options['arch'], subarch=self.options['subarch'])

        if file_in_current_configuration(filename,
                            arch=self.options['arch'], subarch=self.options['subarch']) != "n":
            return_dict['blocks_allyesconfig'] = \
                set(get_conditional_blocks(filename, autoconf_h, configuration_blocks=False)) \
                    | set(["B00"])
        else:
            return_dict['blocks_allyesconfig'] = set()

        return return_dict


def select_framework(identifier, options):

    frameworks = {'kbuild' : KbuildBuildFramework,
                  'linux'  : KbuildBuildFramework,
                  'bare'   : BareBuildFramework}

    if identifier is None:
        identifier='bare'
        try:
            logging.info("Detected Linux version %s, selecting kbuild framework",
                         get_linux_version())
            identifier='linux'
        except NotALinuxTree:
            pass

    if os.environ.has_key('ARCH'):
        options['arch'] = os.environ['ARCH']

        if os.environ.has_key('SUBARCH'):
            options['subarch'] = os.environ['SUBARCH']
        else:
            options['subarch'] = guess_subarch_from_arch(options['arch'])

    if identifier in frameworks:
        bf = frameworks[identifier](options)
    else:
        raise RuntimeError("Build framework '%s' not found" % \
                               options['framework'])
    return bf



# Copyright (C) 2012 Reinhard Tartler <tartler@informatik.uni-erlangen.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import unittest as t

import re
import os.path
import glob

from vamos.tools import execute

def find_blocks(cmd, regex):
    (output, _) = execute(cmd, failok=False)
    blocks = set()
    blocks.add(0)
    for line in output:
        m = re.match(regex, line)
        if m:
            blocks.add(int(m.group(1)))

    return blocks

class testVamos(t.TestCase):
    def check_highest_blockno(self, filename):

        self.assertTrue(os.path.exists(filename))

        b1 = find_blocks("zizler -c %s" % filename, r"^B(\d+)")
        b2 = find_blocks("undertaker -j cpppc %s" % filename,
                              r"^&& \( B(\d+) <->")
        self.assertEqual(max(b1), max(b2),
                         "File %s different highest blockno (%d vs %d)" \
                             % (filename, max(b1), max(b2)))

        self.assertEqual(b1, b2,
                         "zizler and undertaker disagree on found blocks")
        return len(b1)

    def check_blocksets(self, filename):
        b1 = find_blocks("zizler -cC %s" % filename, r"^B(\d+)")
        b2 = find_blocks("undertaker -j cpppc %s" % filename,
                              r"^&& \( B(\d+) <->")
        self.assertEqual(len(b1 - b2), 0)

    def test_highest_blockno(self):
        files = set()
        for f in glob.glob("../undertaker/validation/*.c"):
            self.check_highest_blockno(f)
            files.add(f)
        print "\nhighest blockno: Checked %d files" % len(files)

    def test_blocksets(self):
        files = set()
        for f in glob.glob("../undertaker/validation/*.c"):
            self.check_blocksets(f)
            files.add(f)
        print "\nblocksets: Checked %d files" % len(files)

if __name__ == '__main__':
    t.main()



#!/usr/bin/env python
#
#   vampyr - extracts presence implications from kconfig dumps
#
# Copyright (C) 2011 Christian Dietrich <christian.dietrich@informatik.uni-erlangen.de>
# Copyright (C) 2011 Reinhard Tartler <tartler@informatik.uni-erlangen.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import unittest as t

from vamos.vampyr.Messages import SparseMessage

class TestSparseMessage(t.TestCase):

    def setUp(self):
        self.sparse_output = """\
kernel/fork.c:709: warning: incorrect type in argument 1 (different signedness) | expected unsigned int [noderef] [usertype] <asn:1>*uaddr | got int [noderef] <asn:1>*clear_child_tid
include/linux/mm.h:723: warning: potentially expensive pointer subtraction
kernel/fork.c:85: warning: symbol 'max_threads' was not declared. Should it be static?
kernel/fork.c:210: warning: symbol 'fork_init' was not declared. Should it be static?
"""
        messages = SparseMessage.preprocess_messages(self.sparse_output.splitlines())
        self.messages = [SparseMessage(None, x) for x in messages]

    def test_number_of_messages(self):
        self.assertEqual(len(self.sparse_output.splitlines()),
                         len(self.messages))

    def test_equality(self):
        self.assertNotEqual(self.messages[1], self.messages[2])

if __name__ == '__main__':
    t.main()




from distutils.core import setup

with open('VERSION') as f:
    version = f.read().rstrip('\n')

setup (name='vamos',
       version=version,
       package_dir={'vamos': 'python/vamos'},
       packages=['vamos',
                 'vamos.rsf2model',
                 'vamos.golem',
                 'vamos.vampyr',
                 ],
       scripts=['python/golem',
                'python/rsf2model',
                ],
       )



#
#   vamos - common auxiliary functionality
#
# Copyright (C) 2011 Christian Dietrich <christian.dietrich@informatik.uni-erlangen.de>
# Copyright (C) 2011 Reinhard Tartler <tartler@informatik.uni-erlangen.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import logging
import os.path

from subprocess import *


class CommandFailed(RuntimeError):
    """ Indicates that some command failed

    Attributes:
        command    -- the command that failed
        returncode -- the exitcode of the failed command
    """
    def __init__(self, command, returncode):
        assert(returncode != 0)
        self.command = command
        self.returncode = returncode
        self.repr = "Command %s failed to execute (returncode: %d)" % \
            (command, returncode)
        RuntimeError.__init__(self, self.repr)
    def __str__(self):
        return self.repr


def setup_logging(log_level):
    """ setup the logging module with the given log_level """

    l = logging.WARNING # default
    if log_level == 1:
        l = logging.INFO
    elif log_level >= 2:
        l = logging.DEBUG

    logging.basicConfig(level=l)


def execute(command, echo=True, failok=True):
    """
    executes 'command' in a shell.

    optional parameter echo can be used to suppress emitting the command
    to logging.debug.

    if failok is set to false, an RuntimeError will be raised with the
    full commandname and exitcode.

    returns a tuple with
     1. the command's standard output as list of lines
     2. the exitcode
    """
    os.environ["LC_ALL"] = "C"
    os.environ["LC_MESSAGES"] = "C"

    if echo:
        logging.debug("executing: " + command)
    p = Popen(command, stdout=PIPE, stderr=STDOUT, shell=True)
    (stdout, _)  = p.communicate() # stderr is merged into STDOUT
    if not failok and p.returncode != 0:
        raise CommandFailed(command, p.returncode)
    if len(stdout) > 0 and stdout[-1] == '\n':
        stdout = stdout[:-1]
    return (stdout.__str__().rsplit('\n'), p.returncode)


def calculate_worklist(args, batch_mode=False):
    """
    Calculates a sanitizes worklist from a list of given arguments

    This method is intended to be passed lists from the command line. It
    is passed a list of filenames, which are tested individually for
    existance. Non-existing filenames are skipped with a warning.

    Optionally, the given files can be interpreted as worklist. In this
    case, each line in the given files is tested for existance and
    considered for the worklist.

    In any case, the caller can rely on a set of filenames which are
    good for processing.
    """

    worklist = set()

    if batch_mode:
        for worklist_file in args:
            oldlen = len(worklist)
            try:
                with open(worklist_file) as fd:
                    for line in fd:
                        f = os.path.normpath(line.rstrip())
                        if os.path.exists(f):
                            worklist.add(f)
                        else:
                            logging.warning("Skipping non-existent file %s", f)
            except IOError:
                logging.warning("failed to process worklist %s", worklist_file)
            logging.info("Added %d files from worklist %s",
                         len(worklist) - oldlen, worklist_file)
    else:
        for filename in args:
            f = os.path.normpath(filename)
            if os.path.exists(f):
                worklist.add(f)
            else:
                logging.warning("Skipping non-existent file %s", f)
    logging.info("Processing %d files in total", len(worklist))

    return worklist


def check_tool(tool):
    """
    checks if the specified tool is available and executable

    NB: the specified command will be executed and should therefore not
    have any side effects. For gcc or make pass an option such as '--version'.
    """

    try:
        execute(tool, failok=False)
    except CommandFailed:
        logging.error("Failed to locate %s, check your installation", tool)
        return False

    return True

def get_online_processors():
    """
    tries to determine how many processors are currently online
    """

    threads = 1
    try:
        output, _= execute("getconf _NPROCESSORS_ONLN", failok=False)
        threads = int(output[0])
    except StandardError:
        logging.error("Unable to determine number of online processors, " +\
                          "assuming 1")
    return threads
